#
# Copyright (C) 2021 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
from hypothesis import settings, given, HealthCheck
from hypothesis.strategies import integers

import fwdpy11
import msprime
import numpy as np
import pytest


def generate_msprime_ancestry(
    msprime_seed,
    fp11_seed,
    ndescendants,
    N=100,
    alpha=1000,  # 2Ns
    rho=100,
    L=1.0,
    max_attempts=100000,
):
    for initial_ts in msprime.sim_ancestry(
        samples=N,
        population_size=2 * N,
        recombination_rate=rho / 4 / N,
        random_seed=msprime_seed,
        sequence_length=1.0,
        num_replicates=max_attempts,
    ):
        pop = fwdpy11.DiploidPopulation.create_from_tskit(initial_ts)
        pdict = {
            "recregions": [fwdpy11.PoissonInterval(0, 1, 5e-2)],
            "gvalue": fwdpy11.Multiplicative(2.0),
            "rates": (0, 0, None),
            "demography": fwdpy11.ForwardDemesGraph.tubes([pop.N],
                                                          burnin=10*pop.N,
                                                          burnin_is_exact=True),
            "prune_selected": False,
            "simlen": 10 * pop.N,
        }
        _ = fwdpy11.ModelParams(**pdict)

        rng = fwdpy11.GSLrng(fp11_seed)
        data = fwdpy11.NewMutationData(
            effect_size=alpha / 2 / N, dominance=1.0)
        idx = pop.add_mutation(
            rng, ndescendants=ndescendants, data=data, window=(0.49, 0.51)
        )
        if idx is None:
            continue

        return pop, idx
    return None, None


# This test triggers GitHub issue 836
@given(msprime_seed=integers(1, int(2**32 - 1)),
       fp11_seed=integers(1, int(2**32 - 1)),
       ndescendants=integers(2, 101))
@settings(deadline=None, max_examples=25)
def test_ndescendants(msprime_seed, fp11_seed, ndescendants):
    print(msprime_seed, fp11_seed, ndescendants)
    pop, idx = generate_msprime_ancestry(msprime_seed, fp11_seed, ndescendants)
    if pop is not None:
        _ = pop.dump_tables_to_tskit()


# NOTE: this is copied from test/test_tree_sequences.py
# FIXME: this should be a more general fixture?
@pytest.fixture
def set_up_quant_trait_model():
    # TODO add neutral variants
    N = 1000
    demography = fwdpy11.ForwardDemesGraph.tubes(
        [N], burnin=100, burnin_is_exact=True)
    rho = 2.0
    # theta = 100.
    # nreps = 500
    # mu = theta/(4*N)
    r = rho / (4 * N)
    Opt = fwdpy11.Optimum
    GSSmo = fwdpy11.GaussianStabilizingSelection.single_trait(
        [Opt(when=0, optimum=0.0, VS=1.0), Opt(when=N, optimum=1.0, VS=1.0)]
    )
    a = fwdpy11.Additive(2.0, GSSmo)
    p = {
        "nregions": [],
        "sregions": [fwdpy11.GaussianS(0, 1, 1, 0.25)],
        "recregions": [fwdpy11.PoissonInterval(0, 1, r)],
        "rates": (0.0, 0.025, None),
        "gvalue": a,
        "prune_selected": False,
        "demography": demography,
        "simlen": demography.final_generation
    }
    params = fwdpy11.ModelParams(**p)
    rng = fwdpy11.GSLrng(101 * 45 * 110 * 210)
    pop = fwdpy11.DiploidPopulation(N, 1.0)
    return params, rng, pop


def test_new_mutation_data_bad_construction():
    with pytest.raises(ValueError):
        # Mutation is neutral
        fwdpy11.NewMutationData(effect_size=0.0, dominance=1.0)

    with pytest.raises(ValueError):
        fwdpy11.NewMutationData(
            effect_size=np.nan,
            dominance=1.0,
        )

    with pytest.raises(ValueError):
        fwdpy11.NewMutationData(
            effect_size=-1e-3,
            dominance=np.nan,
        )

    with pytest.raises(ValueError):
        fwdpy11.NewMutationData(
            effect_size=0.0,  # 0
            dominance=np.nan,
            esizes=[0.0] * 4,  # all 0
            heffects=[0.5] * 4,
        )

    with pytest.raises(ValueError):
        fwdpy11.NewMutationData(
            effect_size=1e-3,
            dominance=np.nan,
            esizes=[0.0] * 4,  # length 4
            heffects=[0.5] * 5,  # length 5
        )


@given(ndescendants=integers(1, 11),
       seed=integers(1, int(2**32 - 1)),
       msprime_seed=integers(1, int(2**32 - 1)))
@settings(deadline=None, max_examples=10)
def test_add_mutation_in_single_deme(ndescendants, seed, msprime_seed):
    ts = msprime.sim_ancestry(
        50, population_size=50, sequence_length=10, random_seed=msprime_seed
    )
    pop = fwdpy11.DiploidPopulation.create_from_tskit(ts)
    data = fwdpy11.NewMutationData(effect_size=-1e-3, dominance=1.0)
    rng = fwdpy11.GSLrng(seed)
    key = pop.add_mutation(rng, ndescendants=ndescendants, data=data)
    # Skip over cases where ndescendants could not be satisfied
    if key is not None:
        assert key == 0
        assert pop.mcounts[key] == ndescendants

        count = 0
        for d in pop.diploids:
            for i in [d.first, d.second]:
                assert i >= 0
                assert i < len(pop.haploid_genomes)
                for k in pop.haploid_genomes[i].smutations:
                    if k == key:
                        count += 1
        assert count == pop.mcounts[key]


@given(ndescendants=integers(1, 45),
       msprime_seed=integers(1, int(2**32 - 1)),
       seed=integers(1, int(2**32 - 1)),
       deme=integers(0, 1))
@settings(suppress_health_check=[HealthCheck.function_scoped_fixture],
          deadline=None, max_examples=10)
def test_add_mutation_in_two_deme_model(deme,
                                        ndescendants,
                                        seed,
                                        msprime_seed,
                                        small_split_model):
    demog = msprime.Demography.from_demes(small_split_model)
    ts = msprime.sim_ancestry(
        samples={0: 50, 1: 50},
        random_seed=msprime_seed,
        sequence_length=1.0,
        demography=demog,
        recombination_rate=1,
        discrete_genome=False,
    )
    pop = fwdpy11.DiploidPopulation.create_from_tskit(ts)
    data = fwdpy11.NewMutationData(effect_size=-1e-3, dominance=1.0)
    rng = fwdpy11.GSLrng(seed)
    key = pop.add_mutation(
        rng, ndescendants=ndescendants, deme=deme, data=data)
    # Skip over cases where ndescendants could not be satisfied
    if key is not None:
        assert key == 0
        assert pop.mcounts[key] == ndescendants

        count = 0
        for di, d in enumerate(pop.diploids):
            for i in [d.first, d.second]:
                assert i >= 0
                assert i < len(pop.haploid_genomes)
                for k in pop.haploid_genomes[i].smutations:
                    if k == key:
                        assert pop.diploid_metadata[di].deme == deme
                        count += 1
        assert count == pop.mcounts[key]


@pytest.mark.parametrize("window", [(0.3, 0.6), (0.7, 0.88)])
@given(msprime_seed=integers(1, int(2**32 - 1)),
       fwdpy11_seed=integers(1, int(2**32-1)),
       ndescendants=integers(2, 25),
       deme=integers(0, 1))
@settings(suppress_health_check=[HealthCheck.function_scoped_fixture],
          deadline=None, max_examples=5)
def test_add_mutation_in_two_deme_model_windows(
    deme, ndescendants, small_split_model, window, msprime_seed, fwdpy11_seed
):
    demog = msprime.Demography.from_demes(small_split_model)
    ts = msprime.sim_ancestry(
        samples={0: 50, 1: 50},
        random_seed=msprime_seed,
        sequence_length=1.0,
        demography=demog,
        recombination_rate=1,
        discrete_genome=False,
    )
    pop = fwdpy11.DiploidPopulation.create_from_tskit(ts)
    data = fwdpy11.NewMutationData(effect_size=-1e-3, dominance=1.0)
    rng = fwdpy11.GSLrng(fwdpy11_seed)
    key = pop.add_mutation(
        rng, window=window, ndescendants=ndescendants, deme=deme, data=data
    )
    # Skip over cases where ndescendants could not be satisfied
    if key is not None:
        assert key == 0
        assert pop.mcounts[key] == ndescendants

        pos = pop.mutations[key].pos
        assert pos >= window[0]
        assert pos < window[1]

        count = 0
        for di, d in enumerate(pop.diploids):
            for i in [d.first, d.second]:
                assert i >= 0
                assert i < len(pop.haploid_genomes)
                for k in pop.haploid_genomes[i].smutations:
                    if k == key:
                        assert pop.diploid_metadata[di].deme == deme
                        count += 1
        assert count == pop.mcounts[key]


@pytest.mark.parametrize("ndescendants", [1])
@pytest.mark.parametrize("deme", [0])
def test_add_mutation_in_bad_window(deme, ndescendants, small_split_model):
    demog = msprime.Demography.from_demes(small_split_model)
    ts = msprime.sim_ancestry(
        samples={0: 50, 1: 50},
        random_seed=654321,
        sequence_length=1.0,
        demography=demog,
        recombination_rate=1,
        discrete_genome=False,
    )
    pop = fwdpy11.DiploidPopulation.create_from_tskit(ts)
    data = fwdpy11.NewMutationData(effect_size=0.3, dominance=1.0)
    rng = fwdpy11.GSLrng(1234)
    with pytest.raises(ValueError):
        pop.add_mutation(
            rng, window=(-1, pop.tables.genome_length), ndescendants=1, data=data
        )


def test_adding_to_evolved_pop(set_up_quant_trait_model):
    params, rng, pop = set_up_quant_trait_model
    fwdpy11.evolvets(rng, pop, params, simplification_interval=100)
    assert len(pop.tables.mutations) > 0
    data = fwdpy11.NewMutationData(effect_size=0.3, dominance=1.0)
    key = pop.add_mutation(rng, ndescendants=1, data=data)
    assert key is not None
    assert pop.mcounts[key] == 1
    count = 0
    for d in pop.diploids:
        for i in [d.first, d.second]:
            assert i >= 0
            assert i < len(pop.haploid_genomes)
            for k in pop.haploid_genomes[i].smutations:
                if k == key:
                    count += 1
    assert count == pop.mcounts[key]


def test_attempting_to_add_during_a_simulation(set_up_quant_trait_model):
    params, rng, pop = set_up_quant_trait_model

    class InvalidRecorder(object):
        def __init__(self):
            self.rng = fwdpy11.GSLrng(101)
            self.data = fwdpy11.NewMutationData(effect_size=0.3, dominance=1.0)

        def __call__(self, pop, sampler):
            if pop.generation > 5:
                pop.add_mutation(self.rng, data=self.data)

    with pytest.raises(RuntimeError):
        fwdpy11.evolvets(
            rng, pop, params, simplification_interval=100, recorder=InvalidRecorder()
        )


def test_attempt_to_add_to_pop_without_ancestry():
    rng = fwdpy11.GSLrng(42)
    pop = fwdpy11.DiploidPopulation(100, 1.0)
    data = fwdpy11.NewMutationData(effect_size=0.3, dominance=1.0)
    with pytest.raises(ValueError):
        pop.add_mutation(rng, ndescendants=1, data=data)


@given(msprime_seed=integers(1, int(2**32 - 1)),
       fp11_seed=integers(1, int(2**32-1)),
       ndescendants=integers(1, 300))
@settings(deadline=None, max_examples=25)
def test_add_mutation_to_random_msprime_output(msprime_seed, fp11_seed, ndescendants):
    initial_ts = msprime.sim_ancestry(
        samples=250,
        population_size=500,
        recombination_rate=1e-3,
        random_seed=msprime_seed,
        sequence_length=1.0,
    )
    pop = fwdpy11.DiploidPopulation.create_from_tskit(initial_ts)
    rng = fwdpy11.GSLrng(fp11_seed)
    data = fwdpy11.NewMutationData(effect_size=1e-3, dominance=1.0)
    idx = pop.add_mutation(
        rng, ndescendants=ndescendants, data=data, window=(0.49, 0.51)
    )
    if idx is not None:
        _ = pop.dump_tables_to_tskit()



from typing import List, Tuple, Union

import fwdpy11._fwdpy11
import fwdpy11._types
import numpy as np


def simplify(pop, samples):
    """
    Simplify a TableCollection stored in a Population.

    :param pop: A :class:`fwdpy11.DiploidPopulation`
    :param samples: A list of samples (node indexes).

    :return: The simplified tables and array mapping input sample
             IDs to output IDS

    :rtype: tuple

    Note that the samples argument is agnostic with respect to the time of
    the nodes in the input tables. Thus, you may do things like simplify
    to a set of "currently-alive" nodes plus some or all ancient samples by
    including some node IDs from
    :attr:`fwdpy11.DiploidPopulation.ancient_sample_metadata`.

    If the input contains ancient samples,
    and you wish to include them in the output,
    then you need to include their IDs in the samples argument.

    .. note::

        Due to node ID remapping, the metadata corresponding
        to nodes becomes a bit more difficult to look up.
        You need to use the output ID map, the original IDs, and
        the population's metadata containers.

    .. deprecated:: 0.3.0

        Prefer :func:`fwdpy11.simplify_tables`

    .. versionchanged:: 0.3.0

        Ancient samples are no longer kept by default

    .. versionchanged:: 0.5.0

        No longer requires a :class:`MutationVector` argument.

    """
    import warnings

    warnings.warn(
        "This function is deprecated and will be removed soon."
        "Please use fwdpy11.simplify_tables instead",
        category=FutureWarning, stacklevel=2
    )

    ll_t, idmap = fwdpy11._fwdpy11._simplify(pop, samples)
    return fwdpy11._types.TableCollection(ll_t), idmap


def simplify_tables(
    tables: fwdpy11._types.TableCollection, samples: Union[List, np.ndarray]
) -> Tuple[fwdpy11._types.TableCollection, np.ndarray]:
    """
    Simplify a TableCollection.

    :param pop: A table collection.
    :type pop: :class:`fwdpy11.TableCollection`
    :param samples: list of samples
    :type list: list-like or array-like

    :returns: A simplified TableCollection and an array containing remapped sample ids.
    :rtype: tuple

    .. versionadded:: 0.3.0

    """
    ll_t, idmap = fwdpy11._fwdpy11._simplify_tables(tables, samples)

    return fwdpy11._types.TableCollection(ll_t), idmap



#
# Copyright (C) 2017-2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
import fwdpy11
import fwdpy11.custom_genetic_value_decorators


@fwdpy11.custom_genetic_value_decorators.default_update
@fwdpy11.custom_genetic_value_decorators.genetic_value_noise_default_clone
class PyNoise(fwdpy11.GeneticValueNoise):
    def __init__(self):
        fwdpy11.GeneticValueNoise.__init__(self)

    def __call__(self, data: fwdpy11.DiploidGeneticValueNoiseData) -> float:
        return fwdpy11.gsl_ran_gaussian_ziggurat(data.rng, 0.1)



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

""" fwdpy11

    isort:skip_file
"""
import sys

from fwdpy11._version import version as __version__  # NOQA

from ._dev import *  # NOQA
from ._fwdpy11 import *  # NOQA

from . import discrete_demography  # NOQA

from .regions import *  # NOQA
from .genetic_map_unit import (  # NOQA
    PoissonPoint,
    PoissonInterval,
    BinomialInterval,
    BinomialIntervalMap,
    BinomialPoint,
    FixedCrossovers,
)
from .mutation_dominance import (  # NOQA
    FixedDominance,
    UniformDominance,
    ExponentialDominance,
    LargeEffectExponentiallyRecessive,
)
from .genetic_values import (  # NOQA
    PleiotropicOptima,
    Optimum,
    GaussianStabilizingSelection,
    GSS,
    GSSmo,
    MultivariateGSS,
    MultivariateGSSmo,
    NoNoise,
    GaussianNoise,
    Additive,
    Multiplicative,
    GBR,
    StrictAdditiveMultivariateEffects,
)

from ._types import (  # NOQA
    DataMatrix,
    DataMatrixIterator,
    TableCollection,
    DiploidPopulation,
    TreeIterator,
    VariantIterator,
    NewMutationData,
    ForwardDemesGraph,
    DemographicModelDetails,
    DemographicModelCitation
)  # NOQA
from ._types.demography_debugger import DemographyDebugger  # NOQA
from ._types.model_params import ModelParams, MutationAndRecombinationRates  # NOQA
from ._evolvets import *  # NOQA

from ._functions import (  # NOQA
    data_matrix_from_tables,
    infinite_sites,
    make_data_matrix,
    simplify,
    simplify_tables,
    _validate_regions,
)  # NOQA

if sys.version_info[0] < 3:
    raise ValueError("Python3 required!")



import demes
import fwdpy11
import pytest


@pytest.fixture
def model_triggering_issue_792():
    yaml = """
description: Two deme model where only one makes it to the end.
time_units: generations
demes:
- name: ancestral
  description: ancestral deme, two epochs
  epochs:
  - {end_time: 50, start_size: 50}
- name: deme1
  description: child 1
  epochs:
  - {start_size: 50, end_size: 50, end_time: 10}
  ancestors: [ancestral]
- name: deme2
  description: child 2
  epochs:
  - {start_size: 50, end_size: 50, end_time: 0}
  ancestors: [ancestral]
"""
    g = demes.loads(yaml)
    return g


def test_issue_792(model_triggering_issue_792):
    """
    Test of GitHub Issue 792
    """
    demog = fwdpy11.discrete_demography.from_demes(model_triggering_issue_792, burnin=1)
    pop = fwdpy11.DiploidPopulation(50, 1)

    pdict = {
        "rates": (0, 0, 0),
        "gvalue": fwdpy11.Multiplicative(2.0),
        "simlen": demog.metadata["total_simulation_length"],
        "demography": demog,
    }

    params = fwdpy11.ModelParams(**pdict)

    rng = fwdpy11.GSLrng(100)

    fwdpy11.evolvets(rng, pop, params, 10)

    _ = pop.dump_tables_to_tskit()



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
from hypothesis import settings, given, HealthCheck
from hypothesis.strategies import booleans, integers

import pytest

import fwdpy11


class CountSamplesPerTimePoint(object):
    def __init__(self):
        self.sample_timepoints = []
        self.sample_sizes = []
        self.timepoint_seen = {}

    def __call__(self, pop):
        # Get the most recent ancient samples
        # and record their number.  We do this
        # by a "brute-force" approach
        for t, n, _ in pop.sample_timepoints(False):
            assert t != pop.generation
            if t not in self.timepoint_seen:
                self.timepoint_seen[t] = 1
            else:
                self.timepoint_seen[t] += 1
            if t not in self.sample_timepoints:
                self.sample_timepoints.append(t)
                self.sample_sizes.append(len(n) // 2)

            # simplify to each time point
            tables, idmap = fwdpy11.simplify_tables(pop.tables, n)
            for ni in n:
                assert idmap[ni] != fwdpy11.NULL_NODE
                assert tables.nodes[idmap[ni]].time == t


@pytest.fixture(
    scope="function",
    params=[
        {"nregions": [], "rates": (0.0, 0.025, None)},
        {
            "nregions": [fwdpy11.Region(0.0, 1.0, 1.0)],
            "rates": (0.025, 0.025, None),
        },
    ],
)
def pdict(request):
    """
    Build models w/ and w/o neutral mutations.
    """
    pd = {
        "nregions": request.param["nregions"],
        "sregions": [fwdpy11.ExpS(beg=0, end=1, weight=1, mean=0.2)],
        "recregions": [fwdpy11.PoissonInterval(0, 1, 1e-2)],
        "gvalue": [fwdpy11.Multiplicative(2.0)],
        "rates": request.param["rates"],
        "demography": None,
    }
    return pd


def make_counter():
    return lambda: CountSamplesPerTimePoint()


def make_none():
    return lambda: None


@pytest.fixture(scope="function", params=[make_counter, make_none])
def resetter(request):
    """
    Generate a value for post_simplification_recorder
    """
    return request.param()()


@pytest.fixture(params=[5, 10, 25, 33, 50, 73, 100, 101])
def simlen(request):
    return request.param


@pytest.fixture(params=[1, 2, 5, 10, 33])
def simplification_inteval(request):
    return request.param


@pytest.fixture(params=[True, False])
def prune_selected(request):
    return request.param


@given(popsize=integers(50, 200),
       simlen=integers(5, 102),
       simplification_inteval=integers(1, 50),
       seed=integers(1, int(2**32 - 1)),
       prune_selected=booleans(),
       resetter=booleans())
@settings(suppress_health_check=[HealthCheck.function_scoped_fixture],
          deadline=None, max_examples=5)
def test_ancient_samples_and_neutral_mutations(
    pdict, simlen, simplification_inteval, prune_selected, seed, popsize, resetter
):
    """
    The test involving neutral mutations test GitHub issue 639 and 646
    """
    mslike_pop = fwdpy11.DiploidPopulation(popsize, 1.0)
    demography = fwdpy11.ForwardDemesGraph.tubes(mslike_pop.deme_sizes()[1],
                                                 burnin=simlen,
                                                 burnin_is_exact=True)
    if resetter is False:
        ancient_sample_resetter = None
    else:
        ancient_sample_resetter = CountSamplesPerTimePoint()
    rng = fwdpy11.GSLrng(seed)
    ancient_sample_recorder = fwdpy11.RandomAncientSamples(
        seed=42, samplesize=10, timepoints=[i for i in range(1, simlen + 1)]
    )
    pdict["simlen"] = demography.final_generation
    pdict["prune_selected"] = prune_selected
    pdict["demography"] = demography
    params = fwdpy11.ModelParams(**pdict)
    fwdpy11.evolvets(
        rng,
        mslike_pop,
        params,
        simplification_inteval,
        recorder=ancient_sample_recorder,
        post_simplification_recorder=ancient_sample_resetter,
    )

    if ancient_sample_resetter is not None:
        assert len(mslike_pop.ancient_sample_metadata) == 0

        # FIXME: check that the number of sampled time points is correct?
        for _, j in ancient_sample_resetter.timepoint_seen.items():
            assert j == 1

        assert all(
            [i == 10 for i in ancient_sample_resetter.sample_sizes]) is True
    else:
        assert (
            len(mslike_pop.ancient_sample_metadata)
            == len([i for i in range(1, params.simlen)]) * 10
        )

    ti = fwdpy11.TreeIterator(
        mslike_pop.tables, mslike_pop.alive_nodes, update_samples=True
    )
    for t in ti:
        for m in t.mutations():
            s = t.samples_below(m.node)
            if len(s) == 0:
                assert mslike_pop.mcounts_ancient_samples[m.key] > 0
            else:
                assert len(s) == mslike_pop.mcounts[m.key]
                if prune_selected is True and len(s) == 2 * mslike_pop.N:
                    assert mslike_pop.mcounts_ancient_samples[m.key] > 0



#
# Copyright(C) 2020 Kevin Thornton < krthornt @uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software : you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.If not, see < http: //www.gnu.org/licenses/>.
#

"""
Test written while addressing the following GitHub issues:
388
389
390
"""

import pickle
import tempfile
import unittest

import numpy as np

import fwdpy11


def set_up_quant_trait_model():
    N = 1000
    rho = 1.0
    r = rho / (4 * N)
    Opt = fwdpy11.Optimum
    GSSmo = fwdpy11.GaussianStabilizingSelection.single_trait(
        [Opt(when=0, optimum=0.0, VS=1.0), Opt(
            when=N, optimum=1.0, VS=1.0)]
    )
    demography = fwdpy11.ForwardDemesGraph.tubes(
        [N], 10*N+100, burnin_is_exact=True)
    a = fwdpy11.Additive(2.0, GSSmo)
    p = {
        "nregions": [],
        "sregions": [fwdpy11.GaussianS(0, 1, 1, 0.25)],
        "recregions": [fwdpy11.PoissonInterval(0, 1, r)],
        "rates": (0.0, 0.025, None),
        "gvalue": a,
        "prune_selected": False,
        "demography": demography,
        "simlen": N + 100,
    }
    params = fwdpy11.ModelParams(**p)
    rng = fwdpy11.GSLrng(101 * 45 * 110 * 210)
    pop = fwdpy11.DiploidPopulation(N, 1.0)
    return params, rng, pop


def set_up_two_trait_quant_trait_model():
    N = 1000
    rho = 1.0
    r = rho / (4 * N)

    optima = [
        fwdpy11.PleiotropicOptima(when=0, optima=np.zeros(2), VS=2.0),
        fwdpy11.PleiotropicOptima(
            when=N, optima=np.array([np.sqrt(2.0), 0]), VS=2.0
        ),
    ]
    GSSmo = fwdpy11.GaussianStabilizingSelection.pleiotropy(optima)
    a = fwdpy11.StrictAdditiveMultivariateEffects(2, 0, GSSmo)
    vcov = np.identity(2)
    np.fill_diagonal(vcov, 0.25)
    DES = fwdpy11.MultivariateGaussianEffects(0, 1, 1, vcov)
    demography = fwdpy11.ForwardDemesGraph.tubes(
        [N], N+100, burnin_is_exact=True)
    p = {
        "nregions": [],
        "sregions": [DES],
        "recregions": [fwdpy11.PoissonInterval(0, 1, r)],
        "rates": (0.0, 0.025, None),
        "gvalue": a,
        "prune_selected": False,
        "demography": demography,
        "simlen": N + 100,
    }
    params = fwdpy11.ModelParams(**p)
    rng = fwdpy11.GSLrng(101 * 45 * 110 * 210)
    pop = fwdpy11.DiploidPopulation(N, 1.0)
    return params, rng, pop


class PreserveN(object):
    def __init__(self, start, n):
        self.start = start
        self.n = n

    def __call__(self, pop, recorder):
        if pop.generation >= self.start:
            recorder.assign(np.arange(self.n, dtype=np.int32))


class TestNoPleiotropy(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.params, self.rng, self.pop = set_up_quant_trait_model()
        preserver = PreserveN(self.pop.N, 10)
        fwdpy11.evolvets(
            self.rng, self.pop, self.params, 100, preserver, record_gvalue_matrix=True
        )

    def test_alive_genetic_values(self):
        for i, j in zip(self.pop.genetic_values.flatten(), self.pop.diploid_metadata):
            self.assertEqual(i, j.g)

    def test_alive_genetic_value_reconstruction(self):
        ti = fwdpy11.TreeIterator(
            self.pop.tables, self.pop.alive_nodes, update_samples=True
        )
        gv = np.zeros(2 * self.pop.N)
        for t in ti:
            for m in t.mutations():
                for b in t.samples_below(m.node):
                    gv[b] += self.pop.mutations[m.key].s
        gv = gv.reshape((self.pop.N, 2))
        gv = np.sum(gv, axis=1)
        md = np.array(self.pop.diploid_metadata, copy=False)
        self.assertTrue(np.allclose(gv, md["g"]))

    def test_ancient_sample_genetic_values(self):
        for i, j in zip(
            self.pop.ancient_sample_genetic_values, self.pop.ancient_sample_metadata
        ):
            self.assertEqual(i, j.g)

    def test_ancient_sample_genetic_value_reconstruction(self):
        as_gv = self.pop.ancient_sample_genetic_values
        amd = np.array(self.pop.ancient_sample_metadata, copy=False)
        nt = np.array(self.pop.tables.nodes, copy=False)
        ancient_nodes = amd["nodes"][:, 0]
        ancient_node_times = nt["time"][ancient_nodes]
        for time, n, _ in self.pop.sample_timepoints(include_alive=False):
            w = np.where(ancient_node_times == time)[0]
            gvslice = as_gv[w].flatten()
            ti = fwdpy11.TreeIterator(self.pop.tables, n, update_samples=True)
            gv = np.zeros(len(n))
            node_map = np.array([np.iinfo(np.int32).max]
                                * len(nt), dtype=np.int32)
            for i, j in enumerate(n):
                node_map[j] = i
            for t in ti:
                for m in t.mutations():
                    for b in t.samples_below(m.node):
                        gv[node_map[b]] += self.pop.mutations[m.key].s
            gv = gv.reshape((len(w), 2))
            gv = np.sum(gv, axis=1)
            self.assertTrue(np.allclose(gv, gvslice))

    def test_pickling(self):
        pp = pickle.dumps(self.pop, -1)
        up = pickle.loads(pp)
        gv = self.pop.genetic_values
        upgv = up.genetic_values
        self.assertTrue(np.array_equal(gv, upgv))
        gv = self.pop.ancient_sample_genetic_values
        upgv = up.ancient_sample_genetic_values
        self.assertTrue(np.array_equal(gv, upgv))

    def test_pickle_to_file(self):
        with tempfile.NamedTemporaryFile(delete=False) as tfile:
            self.pop.pickle_to_file(tfile)
            tfile.seek(0)
            pop2 = fwdpy11.DiploidPopulation.load_from_pickle_file(tfile)
        gv = self.pop.genetic_values
        gv2 = pop2.genetic_values
        self.assertTrue(np.array_equal(gv, gv2))
        gv = self.pop.ancient_sample_genetic_values
        gv2 = pop2.ancient_sample_genetic_values
        self.assertTrue(np.array_equal(gv, gv2))


class TestTwoTraitsIsotropy(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        """
        Differs from previous test:
        we only preserve after the
        optimum shifts, which makes
        it simpler to verify fitnesses.
        """
        self.params, self.rng, self.pop = set_up_two_trait_quant_trait_model()
        preserver = PreserveN(self.pop.N + 1, 10)
        fwdpy11.evolvets(
            self.rng, self.pop, self.params, 100, preserver, record_gvalue_matrix=True
        )
        self.zopt = np.sqrt(2.0)
        self.VS = 2.0

    def test_alive_genetic_values_focal_trait(self):
        gv = self.pop.genetic_values
        for i, j in zip(range(gv.shape[0]), self.pop.diploid_metadata):
            self.assertEqual(gv[i, 0], j.g)

    def test_alive_genetic_value_reconstruction(self):
        gv = self.pop.genetic_values
        ti = fwdpy11.TreeIterator(
            self.pop.tables, self.pop.alive_nodes, update_samples=True
        )
        gv0 = np.zeros(2 * self.pop.N)
        gv1 = np.zeros(2 * self.pop.N)
        for t in ti:
            for m in t.mutations():
                for b in t.samples_below(m.node):
                    gv0[b] += self.pop.mutations[m.key].esizes[0]
                    gv1[b] += self.pop.mutations[m.key].esizes[1]
        # Check trait 0
        gv0 = gv0.reshape((self.pop.N, 2))
        gv0 = np.sum(gv0, axis=1)
        self.assertTrue(np.allclose(gv[:, 0], gv0))
        # Check trait 1
        gv1 = gv1.reshape((self.pop.N, 2))
        gv1 = np.sum(gv1, axis=1)
        self.assertTrue(np.allclose(gv[:, 1], gv1))

        # Now, we check fitness
        md = np.array(self.pop.diploid_metadata, copy=False)
        for i, j, w in zip(gv0, gv1, md["w"]):
            d0 = np.power(i - self.zopt, 2.0)
            d1 = np.power(j - 0.0, 2.0)
            self.assertTrue(np.isclose(
                np.exp(-(d0 + d1) / (2.0 * self.VS)), w))

    def test_ancient_sample_genetic_values_focal_trait(self):
        gv = self.pop.ancient_sample_genetic_values
        for i, j in zip(range(gv.shape[0]), self.pop.ancient_sample_metadata):
            self.assertEqual(gv[i, 0], j.g)

    def test_ancient_sample_genetic_value_reconstruction(self):
        as_gv = self.pop.ancient_sample_genetic_values
        amd = np.array(self.pop.ancient_sample_metadata, copy=False)
        nt = np.array(self.pop.tables.nodes, copy=False)
        ancient_nodes = amd["nodes"][:, 0]
        ancient_node_times = nt["time"][ancient_nodes]
        for time, n, md in self.pop.sample_timepoints(include_alive=False):
            w = np.where(ancient_node_times == time)[0]
            gvslice = as_gv[w, :]
            ti = fwdpy11.TreeIterator(self.pop.tables, n, update_samples=True)
            gv0 = np.zeros(len(n))
            gv1 = np.zeros(len(n))
            node_map = np.array([np.iinfo(np.int32).max]
                                * len(nt), dtype=np.int32)
            for i, j in enumerate(n):
                node_map[j] = i
            for t in ti:
                for m in t.mutations():
                    for b in t.samples_below(m.node):
                        gv0[node_map[b]] += self.pop.mutations[m.key].esizes[0]
                        gv1[node_map[b]] += self.pop.mutations[m.key].esizes[1]
            # Check trait 0
            gv0 = gv0.reshape((len(w), 2))
            gv0 = np.sum(gv0, axis=1)
            self.assertTrue(np.allclose(gv0, gvslice[:, 0]))
            # Check trait 1
            gv1 = gv1.reshape((len(w), 2))
            gv1 = np.sum(gv1, axis=1)
            self.assertTrue(np.allclose(gv1, gvslice[:, 1]))

            # Now, we check fitness
            for i, j, w in zip(gv0, gv1, md["w"]):
                d0 = np.power(i - self.zopt, 2.0)
                d1 = np.power(j - 0.0, 2.0)
                self.assertTrue(np.isclose(
                    np.exp(-(d0 + d1) / (2.0 * self.VS)), w))

    def test_pickling(self):
        pp = pickle.dumps(self.pop, -1)
        up = pickle.loads(pp)
        gv = self.pop.genetic_values
        upgv = up.genetic_values
        self.assertTrue(np.array_equal(gv, upgv))
        gv = self.pop.ancient_sample_genetic_values
        upgv = up.ancient_sample_genetic_values
        self.assertTrue(np.array_equal(gv, upgv))

    def test_pickle_to_file(self):
        with tempfile.NamedTemporaryFile(delete=False) as tfile:
            self.pop.pickle_to_file(tfile)
            tfile.seek(0)
            pop2 = fwdpy11.DiploidPopulation.load_from_pickle_file(tfile)
        gv = self.pop.genetic_values
        gv2 = pop2.genetic_values
        self.assertTrue(np.array_equal(gv, gv2))
        gv = self.pop.ancient_sample_genetic_values
        gv2 = pop2.ancient_sample_genetic_values
        self.assertTrue(np.array_equal(gv, gv2))


class TestWithFirstGenerationPreserved(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.params, self.rng, self.pop = set_up_quant_trait_model()
        preserver = PreserveN(self.pop.N, 10)
        fwdpy11.evolvets(
            self.rng,
            self.pop,
            self.params,
            100,
            preserver,
            record_gvalue_matrix=True,
            preserve_first_generation=True,
        )

    def test_alive_genetic_values(self):
        for i, j in zip(self.pop.genetic_values.flatten(), self.pop.diploid_metadata):
            self.assertEqual(i, j.g)

    def test_alive_genetic_value_reconstruction(self):
        ti = fwdpy11.TreeIterator(
            self.pop.tables, self.pop.alive_nodes, update_samples=True
        )
        gv = np.zeros(2 * self.pop.N)
        for t in ti:
            for m in t.mutations():
                for b in t.samples_below(m.node):
                    gv[b] += self.pop.mutations[m.key].s
        gv = gv.reshape((self.pop.N, 2))
        gv = np.sum(gv, axis=1)
        md = np.array(self.pop.diploid_metadata, copy=False)
        self.assertTrue(np.allclose(gv, md["g"]))

    def test_ancient_sample_genetic_values(self):
        for i, j in zip(
            self.pop.ancient_sample_genetic_values,
            self.pop.ancient_sample_metadata
        ):
            self.assertEqual(i, j.g)

    def test_ancient_sample_genetic_value_reconstruction(self):
        as_gv = self.pop.ancient_sample_genetic_values
        amd = np.array(self.pop.ancient_sample_metadata, copy=False)
        nt = np.array(self.pop.tables.nodes, copy=False)
        ancient_nodes = amd["nodes"][:, 0]
        ancient_node_times = nt["time"][ancient_nodes]
        for time, n, _ in self.pop.sample_timepoints(include_alive=False):
            w = np.where(ancient_node_times == time)[0]
            gvslice = as_gv[w].flatten()
            ti = fwdpy11.TreeIterator(self.pop.tables, n, update_samples=True)
            gv = np.zeros(len(n))
            node_map = np.array([np.iinfo(np.int32).max]
                                * len(nt), dtype=np.int32)
            for i, j in enumerate(n):
                node_map[j] = i
            for t in ti:
                for m in t.mutations():
                    for b in t.samples_below(m.node):
                        gv[node_map[b]] += self.pop.mutations[m.key].s
            gv = gv.reshape((len(w), 2))
            gv = np.sum(gv, axis=1)
            self.assertTrue(np.allclose(gv, gvslice))

    def test_pickling(self):
        pp = pickle.dumps(self.pop, -1)
        up = pickle.loads(pp)
        gv = self.pop.genetic_values
        upgv = up.genetic_values
        self.assertTrue(np.array_equal(gv, upgv))
        gv = self.pop.ancient_sample_genetic_values
        upgv = up.ancient_sample_genetic_values
        self.assertTrue(np.array_equal(gv, upgv))

    def test_pickle_to_file(self):
        with tempfile.NamedTemporaryFile(delete=False) as tfile:
            self.pop.pickle_to_file(tfile)
            tfile.seek(0)
            pop2 = fwdpy11.DiploidPopulation.load_from_pickle_file(tfile)
        gv = self.pop.genetic_values
        gv2 = pop2.genetic_values
        self.assertTrue(np.array_equal(gv, gv2))
        gv = self.pop.ancient_sample_genetic_values
        gv2 = pop2.ancient_sample_genetic_values
        self.assertTrue(np.array_equal(gv, gv2))


if __name__ == "__main__":
    unittest.main()



#
# Copyright(C) 2021 Kevin Thornton < krthornt @uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software : you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.If not, see < http: //www.gnu.org/licenses/>.
#

import attr

import fwdpy11._fwdpy11

from .class_decorators import (
    attr_add_asblack,
    attr_class_to_from_dict,
    attr_class_pickle_with_super,
)


@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict
@attr.s(auto_attribs=True, frozen=True, repr_ns="fwdpy11")
class FixedDominance(fwdpy11._fwdpy11.MutationDominance):
    """
    Fixed heterozygous effects.

    :param h: The heterozygous effect of a mutation, or "dominance".
    :type h: float

    .. versionadded:: 0.13.0
    """

    h: float = attr.ib(validator=attr.validators.instance_of(float))

    def __attrs_post_init__(self):
        super(FixedDominance, self).__init__(fwdpy11._fwdpy11._fixed_dominance(self.h))


@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict
@attr.s(auto_attribs=True, frozen=True, repr_ns="fwdpy11")
class ExponentialDominance(fwdpy11._fwdpy11.MutationDominance):
    """
    Exponential distribution of heterozygous effects.

    :param m: The mean of the distribution.
    :type m: float

    .. versionadded:: 0.13.0
    """

    m: float = attr.ib(validator=attr.validators.instance_of(float))

    def __attrs_post_init__(self):
        super(ExponentialDominance, self).__init__(
            fwdpy11._fwdpy11._exponential_dominance(self.m)
        )


@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict
@attr.s(auto_attribs=True, frozen=True, repr_ns="fwdpy11")
class UniformDominance(fwdpy11._fwdpy11.MutationDominance):
    """
    Uniform distribution of heterozygous effects.

    :param lo: The lower bound of the range
    :type lo: float
    :param hi: The upper bound of the range
    :type hi: float

    .. versionadded:: 0.13.0
    """

    lo: float = attr.ib(validator=attr.validators.instance_of(float))
    hi: float = attr.ib(validator=attr.validators.instance_of(float))

    def __attrs_post_init__(self):
        super(UniformDominance, self).__init__(
            fwdpy11._fwdpy11._uniform_dominance(self.lo, self.hi)
        )


@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict
@attr.s(auto_attribs=True, frozen=True, repr_ns="fwdpy11")
class LargeEffectExponentiallyRecessive(fwdpy11._fwdpy11.MutationDominance):
    """
    Large effect mutations are more recessive according to
    the function :math:`y \\times e^{-k|s|}`, where :math:`s` is the
    effect size and :math:`y` is a "scaling" parameter.

    :param k: The "tuning" parameter.
    :type k: float
    :type scaling: Scaling parameter
    :type scaling: float

    .. versionadded:: 0.13.0
    """

    k: float = attr.ib(validator=attr.validators.instance_of(float))
    scaling: float = attr.ib(validator=attr.validators.instance_of(float), default=1.0)

    @k.validator
    def _k_positive(self, attribute, value):
        if value <= 0.0:
            raise ValueError(f"{attribute} must be > 0.0")

    def __attrs_post_init__(self):
        super(LargeEffectExponentiallyRecessive, self).__init__(
            fwdpy11._fwdpy11._large_effect_exponentially_recessive(self.k, self.scaling)
        )



import sys

import numpy as np
import pandas as pd

import fwdpy11

N = int(sys.argv[1])
rho = float(sys.argv[2])

yaml = f"""
time_units: generations
demes:
  - name: alpha
    epochs:
      - start_size: {N}
  - name: beta
    epochs:
      - start_size: {N}
migrations:
  - demes: [alpha, beta]
    rate: 1e-2
"""

demography = fwdpy11.ForwardDemesGraph.from_demes(
    yaml, burnin=10 * N + 200, burnin_is_exact=True
)

moving_optimum_deme_0 = fwdpy11.GaussianStabilizingSelection.single_trait(
    optima=[
        fwdpy11.Optimum(when=0, optimum=0.0, VS=1.0),
        fwdpy11.Optimum(when=10 * N, optimum=1.0, VS=1.0),
    ],
)
moving_optimum_deme_1 = fwdpy11.GaussianStabilizingSelection.single_trait(
    optima=[
        fwdpy11.Optimum(when=0, optimum=0.0, VS=1.0),
        fwdpy11.Optimum(when=10 * N, optimum=-1.0, VS=1.0),
    ],
)

# The covariance matrix for effect sizes.
# The marginal Gaussians will have mean zero and sd = 0.1
# The deviates will be highly positively correlated.
sd = 0.1
covariance_matrix = np.array([0.999] * 4).reshape((2, 2))
np.fill_diagonal(covariance_matrix, 1)
covariance_matrix *= np.power(sd, 2)

pdict = {
    "nregions": [],
    "sregions": [
        # Multivariate Gaussian distribution of effect sizes
        fwdpy11.mvDES(
            fwdpy11.MultivariateGaussianEffects(
                beg=0, end=1, weight=1, h=1, cov_matrix=covariance_matrix
            ),
            # Means of zero for each marginal Gaussian
            np.zeros(2),
        )
    ],
    "recregions": [fwdpy11.PoissonInterval(0, 1, rho / (4 * N))],
    "rates": (0, 1e-3, None),
    "demography": demography,
    "simlen": 10 * N + 200,  # 200 gens past optimum shift
    # Specify one gvalue object per deme
    "gvalue": [
        fwdpy11.Additive(
            ndemes=2,  # Number of demes
            scaling=2,  # 0, 0+sh, 0+2s for AA, Aa, and aa, respectively
            # Mapping of trait (genetic) value to fitness
            gvalue_to_fitness=moving_optimum_deme_0,
        ),
        fwdpy11.Additive(ndemes=2, scaling=2, gvalue_to_fitness=moving_optimum_deme_1),
    ],
}

params = fwdpy11.ModelParams(**pdict)
pop = fwdpy11.DiploidPopulation([N, N], 1.0)
rng = fwdpy11.GSLrng(1010)
fwdpy11.evolvets(rng, pop, params, 100)
assert pop.generation == 10 * N + 200
md = np.array(pop.diploid_metadata, copy=False)
df = pd.DataFrame.from_records(md[["deme", "g", "w"]])
g = df.groupby(["deme"])
print(g.mean())



import attr

import fwdpy11.class_decorators
import ll_snowdrift


@fwdpy11.class_decorators.attr_class_to_from_dict
@attr.s()
class DiploidSnowdrift(ll_snowdrift._ll_DiploidSnowdrift):
    """
    This is the user-facing Python representation of
    a simple snowdrift model.
    """

    b1 = attr.ib()
    b2 = attr.ib()
    c1 = attr.ib()
    c2 = attr.ib()
    slope = attr.ib()
    p0 = attr.ib()

    def __attrs_post_init__(self):
        # Need to initialize the C++ base class
        super(DiploidSnowdrift, self).__init__(
            self.b1, self.b2, self.c1, self.c2, self.slope, self.p0
        )

    def __getstate__(self):
        # asdict is provided by the 2nd class
        # decorator (reading from bottom to top)
        # We pickle this class as a tuple.
        # Element 0 contains the kwargs + values
        # needed for __init__ and we send along
        # the current phenytpes from the base class
        return (self.asdict(), self.phenotypes)

    def __setstate__(self, t):
        # update the representation of the Python/attrs
        # class
        self.__dict__.update(t[0])
        # Failure to properly init the C++ base class
        # will lead to segmentation faults.
        super(DiploidSnowdrift, self).__init__(**t[0])
        # Reset the base class attribute
        self.phenotypes = t[1]



import glob
import os
import platform
import re
import subprocess
import sys
from distutils.version import LooseVersion

from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext

if "--debug" in sys.argv:
    DEBUG_MODE = True
    sys.argv.remove("--debug")
else:
    DEBUG_MODE = False

if "--enable-profiling" in sys.argv:
    ENABLE_PROFILING = True
    sys.argv.remove("--enable-profiling")
else:
    ENABLE_PROFILING = False

if "--skip_tests" in sys.argv:
    SKIP_BUILDING_TESTS = True
    sys.argv.remove("--skip_tests")
else:
    SKIP_BUILDING_TESTS = False

if "--disable_lto" in sys.argv:
    SKIP_LTO = True
    sys.argv.remove("--disable_lto")
else:
    SKIP_LTO = False

if "--cpp17" in sys.argv:
    USE_CPP17 = True
    sys.argv.remove("--cpp17")
else:
    USE_CPP17 = False

# if '--assert' in sys.argv:
#     ASSERT_MODE = True
#     sys.argv.remove('--assert')
# else:
#     ASSERT_MODE = False


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=""):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(["cmake", "--version"])
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build the following extensions: "
                + ", ".join(e.name for e in self.extensions)
            )

        if platform.system() == "Windows":
            cmake_version = LooseVersion(
                re.search(r"version\s*([\d.]+)", out.decode()).group(1)
            )
            if cmake_version < "3.1.0":
                raise RuntimeError("CMake >= 3.1.0 is required on Windows")

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        parent = os.path.abspath(os.path.dirname("."))
        cmake_args = [
            "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=" + extdir,
            "-DCMAKE_ARCHIVE_OUTPUT_DIRECTORY=" + parent + "/lib/",
            "-DCMAKE_BUILD=OFF",
        ]

        cfg = "Debug" if DEBUG_MODE is True else "Release"

        build_args = ["--config", cfg]

        if platform.system() == "Windows":
            cmake_args += [
                "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}".format(cfg.upper(), extdir)
            ]
            if sys.maxsize > 2 ** 32:
                cmake_args += ["-A", "x64"]
            build_args += ["--", "/m"]
        else:
            cmake_args += ["-DCMAKE_BUILD_TYPE=" + cfg]
            build_args += ["--", "-j2"]

        if ENABLE_PROFILING is True:
            cmake_args += ["-DENABLE_PROFILING=ON"]

        env = os.environ.copy()
        env["CXXFLAGS"] = '{} -DVERSION_INFO=\\"{}\\"'.format(
            env.get("CXXFLAGS", ""), self.distribution.get_version()
        )

        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        if SKIP_BUILDING_TESTS is True:
            cmake_args.append("-DBUILD_PYTHON_UNIT_TESTS=OFF")

        if SKIP_LTO is True:
            cmake_args.append("-DDISABLE_LTO=ON")

        if USE_CPP17 is True:
            cmake_args.append("-DUSECPP17=ON")

        subprocess.check_call(
            ["cmake", ext.sourcedir] + cmake_args, cwd=self.build_temp, env=env
        )
        subprocess.check_call(
            ["cmake", "--build", "."] + build_args, cwd=self.build_temp
        )


PKGS = [
    "fwdpy11",
    "fwdpy11.conditional_models",
    "fwdpy11.tskit_tools",
    "fwdpy11._functions",
    "fwdpy11._types",
]

ext_modules = [
    CMakeExtension("fwdpy11._fwdpy11"),
]


# Figure out the headers we need to install:
generated_package_data = {}
for root, dirnames, filenames in os.walk("fwdpy11/headers"):
    if (
        "testsuite" not in root
        and "examples" not in root
        and "python_examples" not in root
    ):
        g = glob.glob(root + "/*.hh")
        if len(g) > 0:
            replace = root.replace("/", ".")
            # If there's a header file, we add the directory as a package
            if replace not in PKGS:
                PKGS.append(replace)
            generated_package_data[replace] = ["*.hh"]
        g = glob.glob(root + "/*.hpp")
        if len(g) > 0:
            replace = root.replace("/", ".")
            # If there's a header file, we add the directory as a package
            if replace not in PKGS:
                PKGS.append(replace)
            try:
                if "*.hpp" not in generated_package_data[replace]:
                    generated_package_data[replace].append("*.hpp")
            except:  # NOQA
                generated_package_data[replace] = ["*.hpp"]
        g = glob.glob(root + "/*.tcc")
        if len(g) > 0:
            replace = root.replace("/", ".")
            # If there's a template implementation file,
            # we add the directory as a package
            if replace not in PKGS:
                PKGS.append(replace)
            try:
                if "*.tcc" not in generated_package_data[replace]:
                    generated_package_data[replace].append("*.tcc")
            except:  # NOQA
                generated_package_data[replace] = ["*.tcc"]

if platform.system() == "Darwin":
    generated_package_data["fwdpy11"] = ["lib*.dylib"]
else:
    generated_package_data["fwdpy11"] = ["lib*.so"]

setup(
    ext_modules=ext_modules,
    cmdclass={"build_ext": CMakeBuild},
    packages=PKGS,
    package_data=generated_package_data,
)



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
import unittest

import numpy as np


class TestPickleDiploidPopulation(unittest.TestCase):
    @classmethod
    def setUp(self):
        from quick_pops import quick_neutral_slocus

        self.pop = quick_neutral_slocus()

    def testPickleDiploidsPy(self):
        import pickle

        pd = pickle.dumps(self.pop.diploids)
        p = pickle.loads(pd)
        for i, j in zip(p, self.pop.diploids):
            self.assertEqual(i, j)

    def testPicklePopPy(self):
        import pickle

        p = pickle.dumps(self.pop, -1)
        pp = pickle.loads(p)
        self.assertEqual(pp, self.pop)

    def testPickleTableCollection(self):
        import pickle

        p = pickle.dumps(self.pop.tables)
        up = pickle.loads(p)
        self.assertEqual(up.genome_length, self.pop.tables.genome_length)


class TestPickleDiploidPopulationTreeSequences(unittest.TestCase):
    @classmethod
    def setUp(self):
        import fwdpy11

        self.N = 1000
        self.rho = 1.0
        self.theta = 100.0
        self.nreps = 500
        self.mu = self.theta / (4 * self.N)
        self.r = self.rho / (4 * self.N)

        self.GSS = fwdpy11.GaussianStabilizingSelection.single_trait(
            [fwdpy11.Optimum(VS=1.0, optimum=0.0, when=0)])
        a = fwdpy11.Additive(2.0, self.GSS)
        demography = fwdpy11.ForwardDemesGraph.tubes([self.N],
                                                     burnin=100,
                                                     burnin_is_exact=True)
        self.p = {
            "nregions": [],
            "sregions": [fwdpy11.GaussianS(0, 1, 1, 0.25)],
            "recregions": [fwdpy11.PoissonInterval(0, 1, self.r)],
            "rates": (0.0, 0.025, None),
            "gvalue": a,
            "prune_selected": False,
            "demography": demography,
            "simlen": demography.final_generation
        }
        self.params = fwdpy11.ModelParams(**self.p)
        self.rng = fwdpy11.GSLrng(101 * 45 * 110 * 210)
        self.pop = fwdpy11.DiploidPopulation(self.N, 1.0)
        fwdpy11.evolvets(self.rng, self.pop, self.params, 100)

    def testPickleNodeTable(self):
        import pickle
        import numpy as np

        pn = pickle.dumps(self.pop.tables.nodes)
        up = pickle.loads(pn)
        nodes = np.array(self.pop.tables.nodes)
        nodes2 = np.array(up)
        self.assertTrue(np.array_equal(nodes, nodes2))

    def testPickleEdgeTable(self):
        import pickle
        import numpy as np

        p = pickle.dumps(self.pop.tables.edges)
        u = pickle.loads(p)
        edges = np.array(self.pop.tables.edges)
        edges2 = np.array(u)
        self.assertTrue(np.array_equal(edges, edges2))

    def testPickleMutationTable(self):
        import pickle
        import numpy as np

        p = pickle.dumps(self.pop.tables.mutations)
        u = pickle.loads(p)
        mutations = np.array(self.pop.tables.mutations)
        mutations2 = np.array(u)
        self.assertTrue(np.array_equal(mutations, mutations2))

    def testPickleTableCollection(self):
        import pickle

        p = pickle.dumps(self.pop.tables)
        u = pickle.loads(p)
        self.assertTrue(u == self.pop.tables)

    def testPickleMetaData(self):
        import pickle
        import numpy as np

        p = pickle.dumps(self.pop.diploid_metadata)
        up = pickle.loads(p)
        md = np.array(self.pop.diploid_metadata)
        md2 = np.array(up)
        self.assertTrue(np.array_equal(md, md2))

    def testPicklingToFile(self):
        import fwdpy11
        import lzma
        import os

        fname = "pickleFileTesting.pickle"
        with lzma.open(fname, "wb") as f:
            self.pop.pickle_to_file(f)
        with lzma.open(fname, "rb") as f:
            pop = fwdpy11.DiploidPopulation.load_from_pickle_file(f)
        self.assertEqual(pop.N, self.pop.N)
        self.assertEqual(pop.generation, self.pop.generation)
        self.assertTrue(pop.diploids == self.pop.diploids)
        self.assertTrue(pop.haploid_genomes == self.pop.haploid_genomes)
        self.assertTrue(pop.mutations == self.pop.mutations)
        self.assertTrue(np.array_equal(pop.mcounts, self.pop.mcounts))
        self.assertTrue(pop.tables == self.pop.tables)
        self.assertTrue(pop == self.pop)
        os.remove(fname)


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import argparse
import sys


def print_includes():
    from ._dev import get_fwdpp_includes
    from ._dev import get_includes

    idirs = [get_includes(), get_fwdpp_includes()]

    print(" ".join("-I" + idir for idir in idirs))


def get_fwdpy11_includes():
    from ._dev import get_includes

    return get_includes()


def get_fwdpp_includes():
    from ._dev import get_fwdpp_includes

    return get_fwdpp_includes()


def print_mako():
    from ._dev import minimal_mako

    print(minimal_mako())


def main():
    parser = argparse.ArgumentParser(prog="python -m fwdpy11")
    parser.add_argument(
        "--includes",
        action="store_true",
        help="Print the CPPFLAGS required to use fwdpy11 and fwdpp headers.",
    )
    parser.add_argument(
        "--fwdpy11_headers", action="store_true", help="Get location of fwdpy11 headers"
    )
    parser.add_argument(
        "--fwdpp_headers", action="store_true", help="Get location of fwdpp headers"
    )
    parser.add_argument(
        "--mako",
        action="store_true",
        help="Print minimal mako header for use with cppimport.",
    )
    args = parser.parse_args()
    if not sys.argv[1:]:
        parser.print_help()
    if args.includes:
        print_includes()
        return
    if args.fwdpy11_headers:
        print(get_fwdpy11_includes())
        return
    if args.fwdpp_headers:
        print(get_fwdpp_includes())
        return
    if args.mako:
        print_mako()
        return


if __name__ == "__main__":
    main()



import typing

import attr

from .._fwdpy11 import ll_NewMutationData


@attr.s(auto_attribs=True, frozen=True, repr_ns="fwdpy11", kw_only=True)
class NewMutationData(ll_NewMutationData):
    """
    Data object for :func:`fwdpy11.DiploidPopulation.add_mutation`

    .. versionadded:: 0.16.0

    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param effect_size: The fixed/univariate effect size.
                        Will fill in :attr:`fwdpy11.Mutation.s`
    :type effect_size: float
    :param dominance: Heterozygous effect of the mutation.
                      Will fill in :attr:`fwdpy11.Mutation.h`
    :type dominance: float
    :param esizes: Data to fill :attr:`fwdpy11.Mutation.esizes`.
                   Default is `None`.
    :type esizes: list[float]
    :param heffects: Data to fill :attr:`fwdpy11.Mutation.heffects`.
                     Default is `None`.
    :type heffects: list[float]
    :param label: Data for :attr:`fwdpy11.Mutation.label`.
                  Default is 0.
    :type label: int
    """

    effect_size: float
    dominance: float
    esizes: typing.Optional[typing.List[float]] = None
    heffects: typing.Optional[typing.List[float]] = None
    label: int = 0

    def __attrs_post_init__(self):
        if self.esizes is None and self.heffects is None:
            super(NewMutationData, self).__init__(
                self.effect_size, self.dominance, [], [], self.label
            )
        else:
            super(NewMutationData, self).__init__(
                self.effect_size, self.dominance, self.esizes, self.heffects, self.label
            )



import unittest

import numpy as np


class test_DiploidPopulation(unittest.TestCase):
    """
    Test coercion of population data
    to numpy arrays
    """

    @classmethod
    def setUpClass(self):
        from quick_pops import quick_nonneutral_slocus

        self.pop = quick_nonneutral_slocus()
        self.muts = self.pop.mutations_ndarray

    def testMutations(self):
        self.assertTrue(isinstance(self.muts, np.ndarray))
        self.assertEqual(self.muts.shape[0], len(self.pop.mutations))
        for i, j in zip(self.muts, self.pop.mutations):
            self.assertEqual(i["neutral"], j.neutral)
            self.assertEqual(i["h"], j.h)
            self.assertEqual(i["pos"], j.pos)
            self.assertEqual(i["g"], j.g)
            self.assertEqual(i["s"], j.s)

    def testDiploidTraits(self):
        dips = np.array(self.pop.diploid_metadata)
        for i, j in zip(dips, self.pop.diploid_metadata):
            self.assertEqual(i["g"], j.g)
            self.assertEqual(i["e"], j.e)
            self.assertEqual(i["w"], j.w)

    def testDiploidTraitsSlice(self):
        dips = np.array(self.pop.diploid_metadata[15:103:11])
        for i, j in zip(dips, self.pop.diploid_metadata[15:103:11]):
            self.assertEqual(i["g"], j.g)
            self.assertEqual(i["e"], j.e)
            self.assertEqual(i["w"], j.w)

    def testDiploidKeys(self):
        keys = np.array(self.pop.diploids)
        for i, j in zip(keys, self.pop.diploids):
            self.assertEqual(i["first"], j.first)
            self.assertEqual(i["second"], j.second)


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2017-2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

from typing import List, Tuple, Union

import numpy as np

from .._fwdpy11 import ll_DataMatrixIterator
from .._types import TableCollection


class DataMatrixIterator(ll_DataMatrixIterator):
    """
    Efficient iteration over genomic windows.

    This class allows efficient traversal across multiple
    genomic intervals.  The class is iterable,
    and encapsulates the data fields of
    :class:`fwdpy11.DataMatrix`, which are accessible
    as properties.

    To initialize:

    :param tables: A table collection
    :type tables: :class:`fwdpy11.TableCollection`
    :param samples: A list of samples
    :type samples: list or numpy.ndarray
    :param intervals: The :math:`[start, stop)` positions of each interval
    :type intervals: list[tuple]
    :param neutral: If True, include neutral variants
    :type neutral: bool
    :param selected: If True, include selected variants
    :type selected: bool
    :param fixations: (False) If True, include fixations in the sample
    :type fixations: bool


    .. versionadded:: 0.4.4

    .. versionchanged:: 0.5.0

        Initialization longer requires :class:`fwdpy11.MutationVector`

    """

    def __init__(
        self,
        tables: TableCollection,
        samples: Union[List[int], np.ndarray],
        intervals: List[Tuple[float, float]],
        neutral: bool,
        selected: bool,
        fixations=False,
    ):
        super(DataMatrixIterator, self).__init__(
            tables, samples, intervals, neutral, selected, fixations
        )

    @property
    def neutral(self) -> np.ndarray:
        """Genotypes at neutral variants"""
        return self._neutral

    @property
    def neutral_keys(self) -> np.ndarray:
        """Indexes of the neutral variants."""
        return self._neutral_keys

    @property
    def neutral_positions(self) -> np.ndarray:
        """Positions of neutral variants."""
        return self._neutral_positions

    @property
    def selected(self) -> np.ndarray:
        """Genotypes at selected variants."""
        return self._selected

    @property
    def selected_keys(self) -> np.ndarray:
        """Indexes of selected variants."""
        return self._selected_keys

    @property
    def selected_positions(self) -> np.ndarray:
        """Positions of selected variants."""
        return self._selected_positions

    def __next__(self):
        return self._ll_next()

    def __iter__(self):
        return self._ll_iter()



#
# Copyright (C) 2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

# This file contains round-trip tests of correctness using
# fwdpy11.mvDES to specify different DES in different demes.
# Low-level tests of fwdpy11.mvDES itself are in test_regions.py

import unittest
import pytest

import demes
import fwdpy11
import numpy as np


def gvalue_multiplicative(pop, ind, scaling):
    g = 1.0
    keys = [k for k in pop.haploid_genomes[pop.diploids[ind].first].smutations]
    keys.extend(
        [k for k in pop.haploid_genomes[pop.diploids[ind].second].smutations])
    keys = np.array(keys, dtype=np.uint32)
    key_counts = np.unique(keys, return_counts=True)

    ind_md = pop.diploid_metadata[ind]
    for i, j in zip(key_counts[0], key_counts[1]):
        if j == 1:
            g *= (
                1.0
                + pop.mutations[i].esizes[ind_md.deme]
                * pop.mutations[i].heffects[ind_md.deme]
            )
        elif j == 2:
            g *= 1.0 + scaling * pop.mutations[i].esizes[ind_md.deme]

    return g


class TestMultiplicativeWithExpSNoMigration(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.pop = fwdpy11.DiploidPopulation([50, 50], 1)
        mvDES = fwdpy11.mvDES(
            [fwdpy11.ExpS(0, 1, 1, 0.1), fwdpy11.ExpS(0, 1, 1, -0.1)],
            np.zeros(2),
            np.identity(2),
        )

        self.pdict = {
            "nregions": [],
            "sregions": [mvDES],
            "recregions": [fwdpy11.PoissonInterval(0, 1, 1e-2)],
            "rates": (0, 1e-2, None),
            "gvalue": fwdpy11.Multiplicative(2.0, ndemes=2),
            "demography": fwdpy11.ForwardDemesGraph.tubes(self.pop.deme_sizes()[1], 1),
            "simlen": 100,
            "prune_selected": True,
        }
        self.params = fwdpy11.ModelParams(**self.pdict)
        self.rng = fwdpy11.GSLrng(918273)
        fwdpy11.evolvets(self.rng, self.pop, self.params, 100)
        n0 = 0
        n1 = 0
        for i, md in enumerate(self.pop.diploid_metadata):
            ng0 = len(
                self.pop.haploid_genomes[self.pop.diploids[i].first].smutations)
            ng1 = len(
                self.pop.haploid_genomes[self.pop.diploids[i].second].smutations)
            if ng0 + ng1 > 0:
                if md.deme == 0:
                    n0 += 1
                elif md.deme == 1:
                    n1 += 1
        if n0 == 0 or n1 == 0:
            self.assertFail(
                "we don't have individuals with mutations in each deme")

    def test_it(self):
        gv = np.zeros(self.pop.N)
        for i in range(self.pop.N):
            gv[i] = gvalue_multiplicative(self.pop, i, 2.0)
        md = np.array(self.pop.diploid_metadata, copy=False)
        self.assertTrue(np.allclose(gv, md["g"]))


class TestGaussianStabilizingSelection(unittest.TestCase):
    """
    Quick sim w/high migration rate
    to put the same mutation in both demes.
    """

    @ classmethod
    def setUpClass(self):
        yaml = """
        time_units: generations
        demes:
         - name: A
           epochs:
            - start_size: 100
         - name: B
           epochs:
            - start_size: 100
        migrations:
         - demes: [A, B]
           rate: 0.1
        """
        demography = fwdpy11.ForwardDemesGraph.from_demes(yaml, 1)
        pdict = {
            "nregions": [],
            "sregions": [
                fwdpy11.mvDES(
                    fwdpy11.MultivariateGaussianEffects(
                        0, 1, 1, h=1, cov_matrix=np.identity(2)
                    ),
                    np.zeros(2),
                )
            ],
            "recregions": [],
            "rates": (0, 1e-3, None),
            "demography": demography,
            "simlen": 100,
            "gvalue": fwdpy11.Additive(
                ndemes=2, scaling=2, gvalue_to_fitness=fwdpy11.GaussianStabilizingSelection.single_trait([fwdpy11.Optimum(optimum=0.0, VS=1.0, when=0)])
            ),
        }

        self.params = fwdpy11.ModelParams(**pdict)
        self.pop = fwdpy11.DiploidPopulation([100, 100], 1.0)
        self.rng = fwdpy11.GSLrng(1010)
        fwdpy11.evolvets(self.rng, self.pop, self.params, 10)

        tv = fwdpy11.TreeIterator(
            self.pop.tables, self.pop.alive_nodes, update_samples=True
        )
        nt = np.array(self.pop.tables.nodes, copy=False)
        demes_with_muts = np.zeros(2)
        for t in tv:
            for m in t.mutations():
                demes = np.unique(nt["deme"][t.samples_below(m.node)])
                for d in demes:
                    demes_with_muts[d] += 1
        assert np.all(demes_with_muts >
                      0), "test requires mutations in both demes"

    def test_genetic_values(self):
        for m in self.pop.diploid_metadata:
            g = 0.0
            for i in [
                self.pop.diploids[m.label].first,
                self.pop.diploids[m.label].second,
            ]:
                for k in self.pop.haploid_genomes[i].smutations:
                    g += self.pop.mutations[k].esizes[m.deme]
            self.assertAlmostEqual(m.g, g)


class TestMultivariateLogNormalS(unittest.TestCase):
    """
    Quick sim w/high migration rate
    to put the same mutation in both demes.
    """

    @ classmethod
    def setUpClass(self):
        yaml = """
        time_units: generations
        demes:
         - name: A
           epochs:
            - start_size: 100
         - name: B
           epochs:
            - start_size: 100
        migrations:
         - demes: [A, B]
           rate: 0.1
        """
        demography = fwdpy11.ForwardDemesGraph.from_demes(yaml, 1)
        ln = fwdpy11.LogNormalS.mv(0, 1, 1, scaling=-200)
        pdict = {
            "nregions": [],
            "sregions": [fwdpy11.mvDES(ln, np.zeros(2), np.identity(2))],
            "recregions": [],
            "rates": (0, 1e-3, None),
            "demography": demography,
            "simlen": 100,
            "gvalue": fwdpy11.Multiplicative(ndemes=2, scaling=2),
        }

        self.params = fwdpy11.ModelParams(**pdict)
        self.pop = fwdpy11.DiploidPopulation([100, 100], 1.0)
        self.rng = fwdpy11.GSLrng(201012)
        fwdpy11.evolvets(self.rng, self.pop, self.params, 10)

        tv = fwdpy11.TreeIterator(
            self.pop.tables, self.pop.alive_nodes, update_samples=True
        )
        nt = np.array(self.pop.tables.nodes, copy=False)
        demes_with_muts = np.zeros(2)
        for t in tv:
            for m in t.mutations():
                demes = np.unique(nt["deme"][t.samples_below(m.node)])
                for d in demes:
                    demes_with_muts[d] += 1
        assert np.all(demes_with_muts >
                      0), "test requires mutations in both demes"

    def test_genetic_values(self):
        for i, m in enumerate(self.pop.diploid_metadata):
            g = gvalue_multiplicative(self.pop, i, 2.0)
            self.assertAlmostEqual(m.g, g)


@ pytest.mark.parametrize("mvDES",
                          [fwdpy11.mvDES(fwdpy11.MultivariateGaussianEffects(0, 1, 1, np.identity(2)), np.zeros(2)),  # NOQA
                           fwdpy11.mvDES(
                              [fwdpy11.ConstantS(0, 1, 1, 0.1),
                               fwdpy11.ConstantS(0, 1, 1, -0.1)],
                              np.zeros(2),
                              np.identity(2),
                          )])
@ pytest.mark.parametrize("gvalue", [fwdpy11.Multiplicative(2., ndemes=2),
                                     fwdpy11.Additive(2., ndemes=2)])
def test_invalid_model(mvDES, gvalue):
    demog = """
description: trigger exception
time_units: generations
demes:
  - name: ancestor
    epochs:
      - start_size: 100
        end_time: 100
  - name: A
    ancestors: [ancestor]
    epochs:
      - start_size: 100
  - name: B
    ancestors: [ancestor]
    epochs:
      - start_size: 100
  - name: C
    ancestors: [ancestor]
    epochs:
      - start_size: 100
    """

    g = demes.loads(demog)
    model = fwdpy11.discrete_demography.from_demes(g, burnin=1)
    pdict = {
        "nregions": [],
        "sregions": [mvDES],
        "recregions": [fwdpy11.PoissonInterval(0, 1, 1e-2)],
        "rates": (0, 1, None),
        "gvalue": gvalue,
        "demography": model,
        "simlen": model.metadata["total_simulation_length"],
        "prune_selected": True,
    }
    params = fwdpy11.ModelParams(**pdict)
    rng = fwdpy11.GSLrng(918273)
    pop = fwdpy11.DiploidPopulation(100, 1.0)
    with pytest.raises(ValueError):
        fwdpy11.evolvets(rng, pop, params, 100)


def test_two_demes_divergent_optima():
    N = 500
    rho = 0

    yaml = f"""
    time_units: generations
    demes:
      - name: alpha
        epochs:
          - start_size: {N}
      - name: beta
        epochs:
          - start_size: {N}
    migrations:
      - demes: [alpha, beta]
        rate: 1e-2
    """

    demography = fwdpy11.ForwardDemesGraph.from_demes(
        yaml, burnin=N + 50, burnin_is_exact=True)

    moving_optimum_deme_0 = fwdpy11.GaussianStabilizingSelection(
        optima=[
            fwdpy11.Optimum(when=0, optimum=0.0, VS=1.0),
            fwdpy11.Optimum(when=N, optimum=1.0, VS=1.0),
        ],
        is_single_trait=True,
    )
    moving_optimum_deme_1 = fwdpy11.GaussianStabilizingSelection(
        optima=[
            fwdpy11.Optimum(when=0, optimum=0.0, VS=1.0),
            fwdpy11.Optimum(when=N, optimum=-1.0, VS=1.0),
            ],
        is_single_trait=True,
    )

    # The covariance matrix for effect sizes.
    # The marginal Gaussians will have mean zero and sd = 0.1
    # The deviates will be highly positively correlated.
    sd = 0.1
    covariance_matrix = np.array([0.999] * 4).reshape((2, 2))
    np.fill_diagonal(covariance_matrix, 1)
    covariance_matrix *= np.power(sd, 2)

    pdict = {
        "nregions": [],
        "sregions": [
            # Multivariate Gaussian distribution of effect sizes
            fwdpy11.mvDES(
                fwdpy11.MultivariateGaussianEffects(
                    beg=0, end=1, weight=1, h=1, cov_matrix=covariance_matrix
                ),
                # Means of zero for each marginal Gaussian
                np.zeros(2),
            )
        ],
        "recregions": [fwdpy11.PoissonInterval(0, 1, rho / (4 * N))],
        "rates": (0, 1e-3, None),
        "demography": demography,
        "simlen": 10 * N + 200,  # 200 gens past optimum shift
        # Specify one gvalue object per deme
        "gvalue": [
            fwdpy11.Additive(
                ndemes=2,  # Number of demes
                scaling=2,  # 0, 0+sh, 0+2s for AA, Aa, and aa, respectively
                # Mapping of trait (genetic) value to fitness
                gvalue_to_fitness=moving_optimum_deme_0,
            ),
            fwdpy11.Additive(ndemes=2, scaling=2,
                             gvalue_to_fitness=moving_optimum_deme_1),
        ],
    }

    params = fwdpy11.ModelParams(**pdict)
    pop = fwdpy11.DiploidPopulation([N, N], 1.0)
    rng = fwdpy11.GSLrng(1010)
    fwdpy11.evolvets(rng, pop, params, 100)


if __name__ == "__main__":
    unittest.main()



import unittest

import fwdpy11


class testAdditive(unittest.TestCase):
    @classmethod
    def setUp(self):
        GN = fwdpy11.GaussianNoise
        self.w = fwdpy11.Additive(2.0)
        self.t = fwdpy11.Additive(2.0, fwdpy11.GaussianStabilizingSelection.single_trait(
            [fwdpy11.Optimum(0.0, 1.0, 0)]))
        self.tn = fwdpy11.Additive(
            1.0, fwdpy11.GaussianStabilizingSelection.single_trait([fwdpy11.Optimum(0.0, 1.0, 0)]), GN(mean=0.1, sd=2.0))

    def testScaling(self):
        self.assertEqual(self.w.scaling, 2.0)
        self.assertEqual(self.t.scaling, 2.0)
        self.assertEqual(self.tn.scaling, 1.0)

    def testFitnessOrTraitBaseClassProperty(self):
        self.assertEqual(self.w.maps_to_fitness, True)
        self.assertEqual(self.w.maps_to_trait_value, False)
        self.assertEqual(self.t.maps_to_fitness, False)
        self.assertEqual(self.t.maps_to_trait_value, True)
        self.assertEqual(self.tn.maps_to_fitness, False)
        self.assertEqual(self.tn.maps_to_trait_value, True)

    def testPickleFitness(self):
        import pickle

        p = pickle.dumps(self.w)
        up = pickle.loads(p)
        self.assertEqual(up.scaling, self.w.scaling)
        self.assertTrue(up.maps_to_fitness)
        self.assertEqual(type(up.noise), type(self.w.noise))
        self.assertEqual(type(up.gvalue_to_fitness),
                         type(self.w.gvalue_to_fitness))

    def testPickleTraitNoNoise(self):
        import pickle

        p = pickle.dumps(self.t)
        up = pickle.loads(p)
        self.assertEqual(up.scaling, self.t.scaling)
        self.assertTrue(up.maps_to_fitness is False)
        self.assertEqual(type(up.noise), type(self.t.noise))
        self.assertEqual(type(up.gvalue_to_fitness),
                         type(self.t.gvalue_to_fitness))

    def testPickleTraitWithNoise(self):
        import pickle

        p = pickle.dumps(self.tn)
        up = pickle.loads(p)
        self.assertEqual(up.scaling, self.tn.scaling)
        self.assertTrue(up.maps_to_fitness is False)
        self.assertEqual(type(up.noise), type(self.tn.noise))
        self.assertEqual(type(up.gvalue_to_fitness),
                         type(self.tn.gvalue_to_fitness))

    def testPickleTraitWithNoiseToFile(self):
        import pickle

        with open("ptest.pickle", "wb") as f:
            pickle.dump(self.tn, f)

        with open("ptest.pickle", "rb") as f:
            up = pickle.load(f)
        self.assertEqual(up.scaling, self.tn.scaling)
        self.assertTrue(up.maps_to_fitness is False)
        self.assertEqual(type(up.noise), type(self.tn.noise))
        self.assertEqual(type(up.gvalue_to_fitness),
                         type(self.tn.gvalue_to_fitness))


class testMultiplicative(unittest.TestCase):
    @ classmethod
    def setUp(self):
        GN = fwdpy11.GaussianNoise
        self.w = fwdpy11.Multiplicative(2.0)
        self.t = fwdpy11.Multiplicative(
            2.0, fwdpy11.GaussianStabilizingSelection.single_trait([fwdpy11.Optimum(0.0, 1.0)]))
        self.tn = fwdpy11.Multiplicative(
            1.0, fwdpy11.GaussianStabilizingSelection.single_trait(
                [fwdpy11.Optimum(0.0, 1.0, 0)]), GN(mean=0.1, sd=2.0)
        )

    def testScaling(self):
        self.assertEqual(self.w.scaling, 2.0)
        self.assertEqual(self.t.scaling, 2.0)
        self.assertEqual(self.tn.scaling, 1.0)

    def testFitnessOrTraitBaseClassProperty(self):
        self.assertEqual(self.w.maps_to_fitness, True)
        self.assertEqual(self.w.maps_to_trait_value, False)
        self.assertEqual(self.t.maps_to_fitness, False)
        self.assertEqual(self.t.maps_to_trait_value, True)
        self.assertEqual(self.tn.maps_to_fitness, False)
        self.assertEqual(self.tn.maps_to_trait_value, True)

    def testPickleFitness(self):
        import pickle

        p = pickle.dumps(self.w)
        up = pickle.loads(p)
        self.assertEqual(up.scaling, self.w.scaling)
        self.assertTrue(up.maps_to_fitness)
        self.assertEqual(type(up.noise), type(self.w.noise))
        self.assertEqual(type(up.gvalue_to_fitness),
                         type(self.w.gvalue_to_fitness))

    def testPickleTraitNoNoise(self):
        import pickle

        p = pickle.dumps(self.t)
        up = pickle.loads(p)
        self.assertEqual(up.scaling, self.t.scaling)
        self.assertTrue(up.maps_to_fitness is False)
        self.assertEqual(type(up.noise), type(self.t.noise))
        self.assertEqual(type(up.gvalue_to_fitness),
                         type(self.t.gvalue_to_fitness))

    def testPickleTraitWithNoise(self):
        import pickle

        p = pickle.dumps(self.tn)
        up = pickle.loads(p)
        self.assertEqual(up.scaling, self.tn.scaling)
        self.assertTrue(up.maps_to_fitness is False)
        self.assertEqual(type(up.noise), type(self.tn.noise))
        self.assertEqual(type(up.gvalue_to_fitness),
                         type(self.tn.gvalue_to_fitness))


class testGBR(unittest.TestCase):
    @ classmethod
    def setUp(self):
        self.gss = fwdpy11.GaussianStabilizingSelection.single_trait(
            [fwdpy11.Optimum(0.0, 1.0, 0)])
        self.gnoise = fwdpy11.GaussianNoise(mean=0.0, sd=1.0)
        self.nonoise = fwdpy11.NoNoise()

    def testPicklingGSS(self):
        import pickle

        gbr = fwdpy11.GBR(self.gss)
        self.assertFalse(gbr.maps_to_fitness)
        self.assertFalse(gbr.maps_to_fitness)
        p = pickle.dumps(gbr, -1)
        up = pickle.loads(p)
        self.assertTrue(up.noise is None)
        self.assertEqual(type(self.gss), type(up.gvalue_to_fitness))

    def testPicklingGSSGaussianNoise(self):
        import pickle

        gbr = fwdpy11.GBR(self.gss, self.gnoise)
        self.assertFalse(gbr.maps_to_fitness)
        self.assertFalse(gbr.maps_to_fitness)
        p = pickle.dumps(gbr, -1)
        up = pickle.loads(p)
        self.assertEqual(type(self.gnoise), type(up.noise))
        self.assertEqual(type(self.gss), type(up.gvalue_to_fitness))


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2017-2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
import math

import attr
import numpy as np

import fwdpy11
import fwdpy11.custom_genetic_value_decorators


@fwdpy11.custom_genetic_value_decorators.default_update
@fwdpy11.custom_genetic_value_decorators.genetic_value_is_trait_default_clone()
@attr.s()
class PyGSS(fwdpy11.GeneticValueIsTrait):
    opt = attr.ib()
    VS = attr.ib()

    def __attrs_post_init__(self):
        fwdpy11.GeneticValueIsTrait.__init__(self)

    def __call__(self, data: fwdpy11.DiploidGeneticValueToFitnessData) -> float:
        return math.e ** (
            -((data.offspring_metadata.g + data.offspring_metadata.e - self.opt) ** 2)
            / (2 * self.VS)
        )


@fwdpy11.custom_genetic_value_decorators.genetic_value_is_trait_default_clone()
@attr.s()
class PyGSSRandomOptimum(fwdpy11.GeneticValueIsTrait):
    opt = attr.ib()
    VS = attr.ib()

    def __attrs_post_init__(self):
        fwdpy11.GeneticValueIsTrait.__init__(self)
        self.optima = []

    def __call__(self, data: fwdpy11.DiploidGeneticValueToFitnessData) -> float:
        return math.e ** (
            -((data.offspring_metadata.g + data.offspring_metadata.e - self.opt) ** 2)
            / (2 * self.VS)
        )

    def update(self, pop: fwdpy11.DiploidPopulation) -> None:
        self.opt = np.random.normal(0.0, 1.0, 1,)[0]
        self.optima.append((pop.generation, self.opt))



import unittest

import fwdpy11
import numpy as np


def set_up_quant_trait_model():
    N = 1000
    rho = 1000.0
    r = rho / (4 * N)

    timepoints = np.array([0], dtype=np.uint32)
    ntraits = 4
    optima = np.array(np.zeros(len(timepoints) * ntraits))
    optima = optima.reshape((len(timepoints), ntraits))
    PO = fwdpy11.PleiotropicOptima
    po = []
    for i, j in enumerate(timepoints):
        po.append(PO(when=int(j), optima=optima[i, :], VS=1.0))
    GSSmo = fwdpy11.GaussianStabilizingSelection.pleiotropy(po)
    cmat = np.identity(ntraits)
    np.fill_diagonal(cmat, 0.1)
    a = fwdpy11.StrictAdditiveMultivariateEffects(ntraits, 0, GSSmo)
    demography = fwdpy11.ForwardDemesGraph.tubes([N], burnin=100,
                                                 burnin_is_exact=True)
    p = {
        "nregions": [],
        "sregions": [fwdpy11.MultivariateGaussianEffects(0, 1, 1, cmat)],
        "recregions": [fwdpy11.PoissonInterval(0, 1, r)],
        "rates": (0.0, 0.001, None),
        "gvalue": a,
        "prune_selected": False,
        "demography": demography,
        "simlen": demography.final_generation
    }
    params = fwdpy11.ModelParams(**p)
    rng = fwdpy11.GSLrng(101 * 45 * 110 * 210)
    pop = fwdpy11.DiploidPopulation(N, 1.0)
    return params, rng, pop, ntraits


class TestMultivariateGSSmo(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.params, self.rng, self.pop, self.ntraits = set_up_quant_trait_model()

    def test_github_issue_310(self):
        fwdpy11.evolvets(self.rng, self.pop, self.params, 500)
        nonmutants = 0
        md = np.array(self.pop.diploid_metadata, copy=False)
        for i, j in enumerate(self.pop.diploids):
            n0 = len(self.pop.haploid_genomes[j.first].smutations)
            n1 = len(self.pop.haploid_genomes[j.second].smutations)
            if md["w"][i] != 1.0:
                self.assertTrue(n0 + n1 > 0)
            elif n1 + n0 == 0:
                nonmutants += 1

            if nonmutants == 0:
                self.fail("Test invalid: zero individuals were mutation-free")


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2021 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

from hypothesis import settings, given
from hypothesis.strategies import booleans, integers

import demes
import fwdpy11
import fwdpy11.conditional_models
import msprime
import pytest

from fwdpy11_test_utilities import seed_list


def make_pop_with_msprime_ancestry(seed):
    N = 100
    rho = 1000
    L = 1.0
    initial_ts = msprime.sim_ancestry(
        samples=N,
        population_size=2 * N,
        recombination_rate=rho / 4 / N,
        random_seed=seed,
        sequence_length=L,
    )
    pop = fwdpy11.DiploidPopulation.create_from_tskit(initial_ts)
    return pop


@pytest.mark.parametrize("count", [-1, 0, 1])
def test_allele_count(count):
    if count <= 0:
        try:
            _ = fwdpy11.conditional_models.AlleleCount(count)
        except ValueError:
            pass
        except Exception as e:
            pytest.fail(f"unexpected exception {e}")
    else:
        _ = fwdpy11.conditional_models.AlleleCount(count)


@pytest.mark.parametrize("r", [(-1, 1), (1, -1), (2, 1)])
def test_bad_allele_count_range(r):
    try:
        _ = fwdpy11.conditional_models.AlleleCountRange(*r)
    except ValueError:
        pass
    except Exception as e:
        pytest.fail(f"unexpected exception {e}")


@pytest.mark.parametrize("r", [(0.0, 0.1), (0.1, 1.0)])
def test_bad_frequency_range(r):
    try:
        _ = fwdpy11.conditional_models.FrequencyRange(*r)
    except ValueError:
        pass
    except Exception as e:
        pytest.fail(f"unexpected exception {e}")


@given(msprime_seed=integers(1, int(2**32-1)),
       fp11_seed=integers(1, int(2**32-1)),
       prune_selected=booleans())
@settings(deadline=None, max_examples=10)
def test_deleterious_mutation_remains_present(msprime_seed, fp11_seed, prune_selected):
    pop = make_pop_with_msprime_ancestry(msprime_seed)
    demography = fwdpy11.ForwardDemesGraph.tubes(pop.deme_sizes()[1],
                                                 burnin=10,
                                                 burnin_is_exact=True)
    pdict = {
        "recregions": [fwdpy11.PoissonInterval(0, 1, 5e-2)],
        "gvalue": fwdpy11.Multiplicative(2.0),
        "rates": (0, 0, None),
        "prune_selected": prune_selected,
        "simlen": 10,
        "demography": demography
    }
    params = fwdpy11.ModelParams(**pdict)
    mutation_data = fwdpy11.conditional_models.NewMutationParameters(
        frequency=fwdpy11.conditional_models.AlleleCount(1),
        data=fwdpy11.NewMutationData(effect_size=-1e-2, dominance=1),
        position=fwdpy11.conditional_models.PositionRange(
            left=0.49, right=0.51),
    )
    rng = fwdpy11.GSLrng(fp11_seed)
    try:
        output = fwdpy11.conditional_models.track_added_mutation(
            rng,
            pop,
            params,
            mutation_data,
            when=0,
            until=10,
        )
        assert output.num_descendant_nodes == 1
        assert output.pop is not None
        assert output.mutation_index is not None
        assert output.mutation_index < len(output.pop.mutations)
        assert output.pop.mutations[output.mutation_index].s == -1e-2
    except fwdpy11.conditional_models.AddMutationFailure:
        pass
    except Exception as e:
        pytest.fail(f"unexpected exception: {e}")


@given(msprime_seed=integers(1, int(2**32-1)),
       fp11_seed=integers(1, int(2**32-1)),
       prune_selected=booleans())
@settings(deadline=None)
def test_deleterious_mutation_remains_present_with_final_recording(
    msprime_seed, fp11_seed, prune_selected
):
    pop = make_pop_with_msprime_ancestry(msprime_seed)
    demography = fwdpy11.ForwardDemesGraph.tubes(pop.deme_sizes()[1],
                                                 burnin=20,
                                                 burnin_is_exact=True)
    pdict = {
        "recregions": [fwdpy11.PoissonInterval(0, 1, 5e-2)],
        "gvalue": fwdpy11.Multiplicative(2.0),
        "rates": (0, 0, None),
        "prune_selected": prune_selected,
        "simlen": 20,
        "demography": demography,
    }
    params = fwdpy11.ModelParams(**pdict)
    mutation_data = fwdpy11.conditional_models.NewMutationParameters(
        frequency=fwdpy11.conditional_models.AlleleCount(1),
        data=fwdpy11.NewMutationData(effect_size=-1e-2, dominance=1),
        position=fwdpy11.conditional_models.PositionRange(
            left=0.49, right=0.51),
    )
    rng = fwdpy11.GSLrng(fp11_seed)
    try:
        output = fwdpy11.conditional_models.track_added_mutation(
            rng,
            pop,
            params,
            mutation_data,
            when=0,
            until=10,
            sampling_policy=fwdpy11.conditional_models.AncientSamplePolicy.COMPLETION,
        )
        assert output.num_descendant_nodes == 1
        assert output.pop.generation == params.simlen
        assert output.pop is not None
        assert output.mutation_index is not None
        assert output.mutation_index < len(output.pop.mutations)
        assert output.pop.mutations[output.mutation_index].s == -1e-2
        assert len(output.pop.ancient_sample_metadata) == pop.N
        for md in output.pop.ancient_sample_metadata:
            for n in md.nodes:
                assert output.pop.tables.nodes[n].time == 10.0
    except fwdpy11.conditional_models.AddMutationFailure as a:
        pytest.fail(f"failed to add singleton variant: {a}")
    except Exception as e:
        pytest.fail(f"unexpected exception: {e}")


@given(msprime_seed=integers(1, int(2**32-1)),
       fp11_seed=integers(1, int(2**32-1)),
       prune_selected=booleans())
@settings(deadline=None)
@pytest.mark.parametrize("alpha", [1000.0])
def test_sweep_from_new_mutation_using_API(msprime_seed, fp11_seed, alpha, prune_selected):
    pop = make_pop_with_msprime_ancestry(msprime_seed)
    demography = fwdpy11.ForwardDemesGraph.tubes(pop.deme_sizes()[1],
                                                 burnin=200,
                                                 burnin_is_exact=True)
    pdict = {
        "recregions": [fwdpy11.PoissonInterval(0, 1, 5e-2)],
        "gvalue": fwdpy11.Multiplicative(2.0),
        "rates": (0, 0, None),
        "prune_selected": prune_selected,
        "simlen": 10 * pop.N,
        "demography": demography
    }
    params = fwdpy11.ModelParams(**pdict)
    mutation_data = fwdpy11.conditional_models.NewMutationParameters(
        frequency=fwdpy11.conditional_models.AlleleCount(1),
        data=fwdpy11.NewMutationData(
            effect_size=alpha / 2 / pop.N, dominance=1),
        position=fwdpy11.conditional_models.PositionRange(
            left=0.49, right=0.51),
    )
    rng = fwdpy11.GSLrng(fp11_seed)
    output = fwdpy11.conditional_models.selective_sweep(
        rng,
        pop,
        params,
        mutation_data,
        fwdpy11.conditional_models.GlobalFixation(),
        sampling_policy=fwdpy11.conditional_models.AncientSamplePolicy.COMPLETION,
    )
    assert output.num_descendant_nodes == 1
    assert output.pop.generation > pop.generation
    if output.pop is not None:
        assert output.pop.mcounts[output.mutation_index] == 2 * output.pop.N
        _ = output.pop.dump_tables_to_tskit()
        if output.pop.fixation_times[output.mutation_index] < output.pop.generation:
            for md in output.pop.ancient_sample_metadata:
                for n in md.nodes:
                    assert (
                        output.pop.tables.nodes[n].time
                        == output.pop.fixation_times[output.mutation_index]
                    )
            ti = fwdpy11.TreeIterator(
                output.pop.tables,
                output.pop.ancient_sample_nodes,
                update_samples=True,
            )
            for t in ti:
                for mutation in t.mutations():
                    assert mutation.key == output.mutation_index
                    assert t.leaf_counts(mutation.node) == 2 * pop.N
                    for s in t.samples_below(mutation.node):
                        assert (
                            output.pop.tables.nodes[s].time
                            == output.pop.fixation_times[output.mutation_index]
                        )

        ti = fwdpy11.TreeIterator(output.pop.tables, output.pop.alive_nodes)
        for t in ti:
            for mutation in t.mutations():
                assert mutation.key == output.mutation_index
                assert t.leaf_counts(mutation.node) == 2 * pop.N


@pytest.mark.parametrize("alpha", [1000.0])
@given(msprime_seed=integers(1, int(2**32-1)),
       fp11_seed=integers(1, int(2**32-1)),
       prune_selected=booleans())
@settings(deadline=None)
def test_sweep_from_new_mutation_using_API_exit_when_finished(
    msprime_seed, fp11_seed, alpha, prune_selected,
):
    pop = make_pop_with_msprime_ancestry(msprime_seed)
    demography = fwdpy11.ForwardDemesGraph.tubes(pop.deme_sizes()[1],
                                                 burnin=200,
                                                 burnin_is_exact=True)
    pdict = {
        "recregions": [fwdpy11.PoissonInterval(0, 1, 5e-2)],
        "gvalue": fwdpy11.Multiplicative(2.0),
        "rates": (0, 0, None),
        "prune_selected": prune_selected,
        "simlen": 10 * pop.N,
        "demography": demography
    }
    params = fwdpy11.ModelParams(**pdict)
    mutation_data = fwdpy11.conditional_models.NewMutationParameters(
        frequency=fwdpy11.conditional_models.AlleleCount(1),
        data=fwdpy11.NewMutationData(
            effect_size=alpha / 2 / pop.N, dominance=1),
        position=fwdpy11.conditional_models.PositionRange(
            left=0.49, right=0.51),
    )
    rng = fwdpy11.GSLrng(fp11_seed)
    output = fwdpy11.conditional_models.selective_sweep(
        rng,
        pop,
        params,
        mutation_data,
        fwdpy11.conditional_models.GlobalFixation(),
        sampling_policy=fwdpy11.conditional_models.AncientSamplePolicy.COMPLETION,
        return_when_stopping_condition_met=True,
    )
    assert len(output.pop.fixation_times) == 1
    assert output.pop.generation == output.pop.fixation_times[0]


@given(msprime_seed=integers(1, int(2**32-1)),
       fp11_seed=integers(1, int(2**32-1)),
       ndescendants=integers(2, 100),
       prune_selected=booleans())
@pytest.mark.parametrize("alpha", [1000.0])
@settings(deadline=None, max_examples=10)
def test_sweep_from_standing_variation_using_API(
    msprime_seed, fp11_seed, ndescendants, alpha, prune_selected
):
    pop = make_pop_with_msprime_ancestry(msprime_seed)
    demography = fwdpy11.ForwardDemesGraph.tubes(pop.deme_sizes()[1],
                                                 burnin=200,
                                                 burnin_is_exact=True)
    pdict = {
        "recregions": [fwdpy11.PoissonInterval(0, 1, 5e-2)],
        "gvalue": fwdpy11.Multiplicative(2.0),
        "rates": (0, 0, None),
        "prune_selected": prune_selected,
        "simlen": 10 * pop.N,
        "demography": demography
    }
    params = fwdpy11.ModelParams(**pdict)
    mutation_data = fwdpy11.conditional_models.NewMutationParameters(
        frequency=fwdpy11.conditional_models.AlleleCount(ndescendants),
        data=fwdpy11.NewMutationData(
            effect_size=alpha / 2 / pop.N, dominance=1),
        position=fwdpy11.conditional_models.PositionRange(
            left=0.49, right=0.51),
    )
    rng = fwdpy11.GSLrng(fp11_seed)
    try:
        output = fwdpy11.conditional_models.selective_sweep(
            rng,
            pop,
            params,
            mutation_data,
            fwdpy11.conditional_models.GlobalFixation(),
        )
        assert output.num_descendant_nodes == ndescendants
        assert output.pop.generation > pop.generation
        if output.pop is not None:
            if output.mutation_index is not None:
                assert output.pop.mcounts[output.mutation_index] == 2 * \
                    output.pop.N
            else:
                assert output.fixation_index is not None
                assert output.fixation_index < len(output.pop.fixations)
                assert output.fixation_index < len(output.pop.fixation_times)
            _ = output.pop.dump_tables_to_tskit()
    except fwdpy11.conditional_models.AddMutationFailure:
        pass
    except Exception as e:
        pytest.fail(f"unexpected exception: {e}")


def count_mutation(pop, idx, focal_deme):
    count = 0
    if focal_deme is None:
        count = pop.mcounts[idx]
    else:
        for md in pop.diploid_metadata:
            if md.deme == focal_deme:
                for genome in [
                    pop.diploids[md.label].first,
                    pop.diploids[md.label].second,
                ]:
                    if idx in pop.haploid_genomes[genome].smutations:
                        count += 1

    return count


def base_demes_model():
    yaml = """
    time_units: generations
    demes:
    - name: ancestor
      epochs:
        - {end_time: 100, start_size: 100}"""
    return yaml


def single_deme_model():
    yaml = """
    - name: derived
      ancestors: [ancestor]
      epochs:
        - {end_time: 0, start_size: 100}
    """
    return base_demes_model() + yaml


def two_demes_with_migration():
    yaml = """
    - name: derived0
      ancestors: [ancestor]
      epochs:
        - {end_time: 0, start_size: 50}
    - name: derived1
      ancestors: [ancestor]
      epochs:
        - {end_time: 0, start_size: 50}
    migrations:
      - demes: [derived0, derived1]
        rate: 0.005
    """
    return base_demes_model() + yaml


def two_demes_with_migration_and_growth():
    yaml = """
    - name: derived0
      ancestors: [ancestor]
      epochs:
        - {end_time: 0, start_size: 50, end_size: 60}
    - name: derived1
      ancestors: [ancestor]
      epochs:
        - {end_time: 0, start_size: 50, end_size: 55}
    migrations:
      - demes: [derived0, derived1]
        rate: 0.005
    """
    return base_demes_model() + yaml


def two_demes_with_no_migration():
    yaml = """
    - name: derived0
      ancestors: [ancestor]
      epochs:
        - {end_time: 0, start_size: 50}
    - name: derived1
      ancestors: [ancestor]
      epochs:
        - {end_time: 0, start_size: 50}
    """
    return base_demes_model() + yaml


@pytest.mark.parametrize("fp11_seed", seed_list(5130125, 5))
@pytest.mark.parametrize("demes_yaml", [single_deme_model()])
@pytest.mark.parametrize("alpha", [1000.0])
@pytest.mark.parametrize("prune_selected", [True, False])
def test_sweep_from_new_mutation_in_single_deme_using_API(fp11_seed, demes_yaml, alpha, prune_selected):
    g = demes.loads(demes_yaml)
    model = fwdpy11.discrete_demography.from_demes(g, burnin=1)
    isizes = model.initial_sizes_list
    pop = fwdpy11.DiploidPopulation(isizes, 1.0)
    pdict = {
        "recregions": [fwdpy11.PoissonInterval(0, 1, 5e-2)],
        "gvalue": fwdpy11.Multiplicative(2.0),
        "rates": (0, 0, None),
        "demography": model,
        "prune_selected": prune_selected,
        "simlen": model.metadata["total_simulation_length"],
    }
    params = fwdpy11.ModelParams(**pdict)
    mutation_data = fwdpy11.conditional_models.NewMutationParameters(
        frequency=fwdpy11.conditional_models.AlleleCount(1),
        data=fwdpy11.NewMutationData(
            effect_size=alpha / 2 / pop.N, dominance=1),
        position=fwdpy11.conditional_models.PositionRange(
            left=0.49, right=0.51),
    )
    rng = fwdpy11.GSLrng(fp11_seed)
    output = fwdpy11.conditional_models.selective_sweep(
        rng,
        pop,
        params,
        mutation_data,
        fwdpy11.conditional_models.GlobalFixation(),
        when=model.metadata["burnin_time"] + 50,
    )
    assert output.num_descendant_nodes == 1
    assert output.pop.generation > pop.generation
    if output.pop is not None:
        if output.mutation_index is not None:
            assert (
                output.pop.mutations[output.mutation_index].g ==
                model.metadata["burnin_time"] + 50
            )
            assert output.pop.mcounts[output.mutation_index] == 2 * output.pop.N
        else:
            assert (
                output.pop.fixations[output.fixation_index].g ==
                model.metadata["burnin_time"] + 50
            )
        _ = output.pop.dump_tables_to_tskit()
    else:
        pytest.fail("test failed")


@pytest.mark.parametrize("fp11_seed", seed_list(5130125, 5))
@pytest.mark.parametrize("demes_yaml", [two_demes_with_migration()])
@pytest.mark.parametrize("alpha", [1000.0])
@pytest.mark.parametrize("prune_selected", [True, False])
def test_sweep_from_new_mutation_with_demography_using_API(
    fp11_seed, demes_yaml, alpha, prune_selected
):
    g = demes.loads(demes_yaml)
    model = fwdpy11.discrete_demography.from_demes(g, burnin=1)
    isizes = model.initial_sizes_list
    pop = fwdpy11.DiploidPopulation(isizes, 1.0)
    pdict = {
        "recregions": [fwdpy11.PoissonInterval(0, 1, 5e-2)],
        "gvalue": fwdpy11.Multiplicative(2.0),
        "rates": (0, 0, None),
        "demography": model,
        "prune_selected": prune_selected,
        "simlen": model.metadata["total_simulation_length"],
    }
    params = fwdpy11.ModelParams(**pdict)
    mutation_data = fwdpy11.conditional_models.NewMutationParameters(
        frequency=fwdpy11.conditional_models.AlleleCount(1),
        data=fwdpy11.NewMutationData(
            effect_size=alpha / 2 / pop.N, dominance=1),
        position=fwdpy11.conditional_models.PositionRange(
            left=0.49, right=0.51),
    )
    rng = fwdpy11.GSLrng(fp11_seed)
    output = fwdpy11.conditional_models.selective_sweep(
        rng,
        pop,
        params,
        mutation_data,
        fwdpy11.conditional_models.GlobalFixation(),
        when=model.metadata["burnin_time"] + 50,
    )
    assert output.num_descendant_nodes == 1
    assert output.pop.generation > pop.generation
    assert output.pop.generation == params.simlen
    if output.pop is not None:
        if output.mutation_index is not None:
            assert (
                output.pop.mutations[output.mutation_index].g ==
                model.metadata["burnin_time"] + 50
            )
            assert output.pop.mcounts[output.mutation_index] == 2 * output.pop.N
        else:
            assert (
                output.pop.fixations[output.fixation_index].g ==
                model.metadata["burnin_time"] + 50
            )
        _ = output.pop.dump_tables_to_tskit()
    else:
        pytest.fail("test failed")


@pytest.mark.parametrize("fp11_seed", seed_list(5130125, 1))
@pytest.mark.parametrize("demes_yaml", [two_demes_with_migration()])
@pytest.mark.parametrize("alpha", [1000.0])
@pytest.mark.parametrize("prune_selected", [True, False])
def test_origination_deme1_fixation_in_deme_2(fp11_seed, demes_yaml, alpha, prune_selected):
    g = demes.loads(demes_yaml)
    model = fwdpy11.discrete_demography.from_demes(g, burnin=1)
    isizes = model.initial_sizes_list
    pop = fwdpy11.DiploidPopulation(isizes, 1.0)
    pdict = {
        "recregions": [fwdpy11.PoissonInterval(0, 1, 5e-2)],
        "gvalue": fwdpy11.Multiplicative(2.0),
        "demography": model,
        "rates": (0, 0, None),
        "prune_selected": prune_selected,
        "simlen": model.metadata["total_simulation_length"],
    }
    params = fwdpy11.ModelParams(**pdict)
    mutation_data = fwdpy11.conditional_models.NewMutationParameters(
        deme=1,
        frequency=fwdpy11.conditional_models.AlleleCount(1),
        data=fwdpy11.NewMutationData(
            effect_size=alpha / 2 / pop.N, dominance=1),
        position=fwdpy11.conditional_models.PositionRange(
            left=0.49, right=0.51),
    )
    rng = fwdpy11.GSLrng(fp11_seed)
    output = fwdpy11.conditional_models.selective_sweep(
        rng,
        pop,
        params,
        mutation_data,
        fwdpy11.conditional_models.FocalDemeFixation(deme=2),
        when=model.metadata["burnin_time"] + 50,
    )
    assert output.num_descendant_nodes == 1
    assert output.pop.generation > pop.generation
    if output.pop is not None:
        if output.mutation_index is not None:
            assert (
                output.pop.mutations[output.mutation_index].g ==
                model.metadata["burnin_time"] + 50
            )
            assert (
                count_mutation(output.pop, output.mutation_index, 2)
                == 2 * output.pop.deme_sizes(as_dict=True)[2]
            )
        else:
            assert (
                output.pop.fixations[output.fixation_index].g ==
                model.metadata["burnin_time"] + 50
            )
        _ = output.pop.dump_tables_to_tskit()


@pytest.mark.parametrize("fp11_seed", seed_list(5130125, 1))
@pytest.mark.parametrize("demes_yaml", [two_demes_with_migration_and_growth()])
@pytest.mark.parametrize("alpha", [1000.0])
@pytest.mark.parametrize("prune_selected", [True, False])
def test_origination_deme1_fixation_in_deme_2_with_growth(fp11_seed, demes_yaml, alpha, prune_selected):
    g = demes.loads(demes_yaml)
    model = fwdpy11.discrete_demography.from_demes(g, burnin=1)
    isizes = model.initial_sizes_list
    pop = fwdpy11.DiploidPopulation(isizes, 1.0)
    pdict = {
        "recregions": [fwdpy11.PoissonInterval(0, 1, 5e-2)],
        "gvalue": fwdpy11.Multiplicative(2.0),
        "demography": model,
        "rates": (0, 0, None),
        "prune_selected": prune_selected,
        "simlen": model.metadata["total_simulation_length"],
    }
    params = fwdpy11.ModelParams(**pdict)
    mutation_data = fwdpy11.conditional_models.NewMutationParameters(
        deme=1,
        frequency=fwdpy11.conditional_models.AlleleCount(1),
        data=fwdpy11.NewMutationData(
            effect_size=alpha / 2 / pop.N, dominance=1),
        position=fwdpy11.conditional_models.PositionRange(
            left=0.49, right=0.51),
    )
    rng = fwdpy11.GSLrng(fp11_seed)
    output = fwdpy11.conditional_models.selective_sweep(
        rng,
        pop,
        params,
        mutation_data,
        fwdpy11.conditional_models.FocalDemeFixation(deme=2),
        when=model.metadata["burnin_time"] + 50,
    )
    assert output.num_descendant_nodes == 1
    assert output.pop.generation > pop.generation
    del pop
    if output.pop is not None:
        if output.mutation_index is not None:
            assert (
                output.pop.mutations[output.mutation_index].g ==
                model.metadata["burnin_time"] + 50
            )

            deme_sizes = output.pop.deme_sizes(as_dict=True)

            for deme, n in zip([1, 2], [60, 55]):
                assert deme_sizes[deme] == n

            assert (
                count_mutation(output.pop, output.mutation_index, 2)
                == 2 * output.pop.deme_sizes(as_dict=True)[2]
            )
        else:
            assert output.fixation_index is not None
            assert (
                output.pop.fixations[output.fixation_index].g ==
                model.metadata["burnin_time"] + 50
            )
        _ = output.pop.dump_tables_to_tskit()
    else:
        pytest.fail("mutation did not fix?")


@pytest.mark.parametrize("fp11_seed", seed_list(5130125, 1))
@pytest.mark.parametrize("demes_yaml", [two_demes_with_no_migration()])
@pytest.mark.parametrize("alpha", [1000.0])
@pytest.mark.parametrize("prune_selected", [True, False])
def test_origination_deme2_fixation_in_deme_2_no_migration(
    fp11_seed, demes_yaml, alpha, prune_selected
):
    g = demes.loads(demes_yaml)
    model = fwdpy11.discrete_demography.from_demes(g, burnin=1)
    isizes = model.initial_sizes_list
    pop = fwdpy11.DiploidPopulation(isizes, 1.0)
    pdict = {
        "recregions": [fwdpy11.PoissonInterval(0, 1, 5e-2)],
        "gvalue": fwdpy11.Multiplicative(2.0),
        "demography": model,
        "rates": (0, 0, None),
        "prune_selected": prune_selected,
        "simlen": model.metadata["total_simulation_length"],
    }
    params = fwdpy11.ModelParams(**pdict)
    mutation_data = fwdpy11.conditional_models.NewMutationParameters(
        deme=2,
        frequency=fwdpy11.conditional_models.AlleleCount(1),
        data=fwdpy11.NewMutationData(
            effect_size=alpha / 2 / pop.N, dominance=1),
        position=fwdpy11.conditional_models.PositionRange(
            left=0.49, right=0.51),
    )
    rng = fwdpy11.GSLrng(fp11_seed)
    output = fwdpy11.conditional_models.selective_sweep(
        rng,
        pop,
        params,
        mutation_data,
        fwdpy11.conditional_models.FocalDemeFixation(deme=2),
        when=model.metadata["burnin_time"] + 50,
    )
    assert output.num_descendant_nodes == 1
    assert output.pop.generation > pop.generation
    if output.pop is not None:
        assert (
            output.pop.mutations[output.mutation_index].g ==
            model.metadata["burnin_time"] + 50
        )
        deme_sizes = output.pop.deme_sizes(as_dict=True)
        assert count_mutation(
            output.pop, output.mutation_index, 2) == 2 * deme_sizes[2]
        assert count_mutation(output.pop, output.mutation_index, 1) == 0
        _ = output.pop.dump_tables_to_tskit()


@pytest.mark.parametrize("seed", seed_list(1512512, 5))
@pytest.mark.parametrize("when", [0, 1, 3])
def test_github_issue_1093(seed, when):
    def setup(prune_selected=False):
        # Dropping mutations requires existing
        # ancestry, which we can get either
        # from a burn-in or from msprime.
        initial_ts = msprime.sim_ancestry(
            samples=500,
            population_size=500,
            recombination_rate=1e-1,
            random_seed=43215,
            sequence_length=1.0,
        )

        # Build the pop from msprime output
        pop = fwdpy11.DiploidPopulation.create_from_tskit(initial_ts)
        demography = fwdpy11.ForwardDemesGraph.tubes(pop.deme_sizes()[1],
                                                     burnin=200,
                                                     burnin_is_exact=True)

        # Set up basic model parameters
        pdict = {
            "recregions": [],
            # Here, 2 means that fitness is multiplicative
            # over 1, 1+hs, 1+2s.
            "gvalue": fwdpy11.Multiplicative(2.0),
            "rates": (0, 0, 0.0),
            "prune_selected": False,
            "simlen": 200,
            "demography": demography
        }
        params = fwdpy11.ModelParams(**pdict)

        return pop, params

    pop, params = setup()
    assert pop.N == 500
    ALPHA = -10.0
    rng = fwdpy11.GSLrng(seed)

    mutation_data = fwdpy11.conditional_models.NewMutationParameters(
        frequency=fwdpy11.conditional_models.AlleleCount(1),
        data=fwdpy11.NewMutationData(
            effect_size=ALPHA / 2 / pop.N, dominance=1),
        position=fwdpy11.conditional_models.PositionRange(
            left=0.49, right=0.51),
    )

    _ = fwdpy11.conditional_models.track_added_mutation(
        rng,
        pop,
        params,
        mutation_data,
        when=when,
        until=7,
        sampling_policy=fwdpy11.conditional_models.AncientSamplePolicy.DURATION,
    )



# Copyright (C) 2023 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
import pytest

import fwdpy11
import numpy as np
import tskit


# NOTE: this is redundant with
# internal checks in the C++ back end.
# However, this code makes use of
# different code paths on the C++ side.
def validate_mutation_table(pop):
    # assert(all([i.time > 0 for i in pop.tables.nodes]))
    ti = fwdpy11.TreeIterator(pop.tables, [i for i in range(2*pop.N)])

    for t in ti:
        for m in t.mutations():
            g = pop.mutations[m.key].g
            nt = pop.tables.nodes[m.node].time
            # Internal mutation  times
            # are meant to be "forwards in time",
            # so a variant must always have an origin
            # time <= that of its node time
            assert g <= nt
            if t.parent(m.node) >= 0:
                pt = pop.tables.nodes[t.parent(m.node)].time
                assert g > pt


def pop_rng_params():
    # Set up parameters
    r = 1e-8
    L = 1e8
    # 1 chromosome, of 1 Morgan length
    assert r * L == 1

    N0 = 1000
    VS = 1.0
    mu = 2.5e-3
    optimum = 0.0

    SD = 0.02
    sregions = fwdpy11.GaussianS(0, L, 1, SD)

    simlen = 3

    pop = fwdpy11.DiploidPopulation(N0, L)

    pdict = {
        "nregions": [],
        "sregions": [sregions],
        "recregions": [fwdpy11.BinomialInterval(0, L, 1)],
        "rates": (0.0, mu, None),
        "gvalue": fwdpy11.Additive(
            scaling=2, gvalue_to_fitness=fwdpy11.GaussianStabilizingSelection.single_trait([fwdpy11.Optimum(optimum=optimum, VS=VS, when=0)])
        ),
        "simlen": simlen,
        "prune_selected": False,
        "demography": fwdpy11.ForwardDemesGraph.tubes(pop.deme_sizes()[1],
                                                      burnin=30,
                                                      burnin_is_exact=True)
    }
    params = fwdpy11.ModelParams(**pdict)

    rng = fwdpy11.GSLrng(424242)

    return pop, rng, params


# Example from Aaron Ragsdale.
# Modified to run much faster
@pytest.fixture
def pop():
    pop, rng, params = pop_rng_params()

    fwdpy11.evolvets(rng, pop, params, 100, suppress_table_indexing=True)

    return pop


@pytest.fixture
def pop_with_ancient_samples():
    pop, rng, params = pop_rng_params()

    def preserver(pop, sampler):
        sampler.assign(np.arange(pop.N, dtype=np.uint32))

    fwdpy11.evolvets(rng, pop, params, 100, recorder=preserver,
                     suppress_table_indexing=True)

    return pop


# Related to GitHub issue 1109.
def test_recreate_pop_from_own_treeseq(pop):
    gen = pop.generation
    ts = pop.dump_tables_to_tskit()
    nm = ts.num_mutations
    assert nm > 0

    pop = fwdpy11.DiploidPopulation.create_from_tskit(
        ts, import_mutations=True)
    assert pop.generation == gen
    assert len(pop.tables.mutations) == nm
    validate_mutation_table(pop)
    assert (all([pop.mutations[i.key].g >= 0 for i in pop.tables.mutations]))  # NOQA


def test_rebuild_individual_metadata(pop):
    md = np.array(pop.diploid_metadata)
    ts = pop.dump_tables_to_tskit()
    pop2 = fwdpy11.DiploidPopulation.create_from_tskit(
        ts, import_individuals=True)
    md2 = np.array(pop2.diploid_metadata)
    assert len(md) == len(md2)
    assert np.array_equal(md['w'], md2['w'])


def test_rebuild_individual_metadata_with_ancient_samples(pop_with_ancient_samples):
    md = np.array(pop_with_ancient_samples.diploid_metadata)
    amd = np.array(pop_with_ancient_samples.ancient_sample_metadata)
    assert len(amd) > 0
    ts = pop_with_ancient_samples.dump_tables_to_tskit()
    pop2 = fwdpy11.DiploidPopulation.create_from_tskit(
        ts, import_individuals=True)
    md2 = np.array(pop2.diploid_metadata)
    assert len(md) == len(md2)
    assert np.array_equal(md['w'], md2['w'])
    amd2 = np.array(pop2.ancient_sample_metadata)
    assert len(amd) == len(amd2)
    assert np.array_equal(amd['w'], amd2['w'])


def test_multiple_export_recreate_evolve_steps():
    pop, rng, params = pop_rng_params()
    for i in range(5):
        fwdpy11.evolvets(rng, pop, params, 100, suppress_table_indexing=True)
        ts = pop.dump_tables_to_tskit()
        pop = fwdpy11.DiploidPopulation.create_from_tskit(
            ts, import_mutations=True)
        assert (all([pop.mutations[i.key].g >= 0 for i in pop.tables.mutations]))  # NOQA
    assert pop.generation == 15


def test_start_pop_from_treeseq_without_mutations(pop):
    ts = pop.dump_tables_to_tskit()
    tables = ts.tables
    nm = len(tables.mutations)
    tables.mutations.clear()
    tables.sites.clear()
    assert len(tables.mutations) == 0
    assert len(tables.sites) == 0
    ts = tables.tree_sequence()

    t = ts.first()
    nodes = [i for i in t.nodes() if t.parent(i) != tskit.NULL and i >
             2000 and (ts.node(t.parent(i)).time - ts.node(i).time) > 1]

    np.random.seed(42)
    rnodes = np.random.choice(nodes, size=nm, replace=False)
    sites = set()
    for n in rnodes:
        time = ts.node(n).time
        ptime = ts.node(t.parent(n)).time
        pos = np.random.uniform(t.interval.left, t.interval.right)
        while pos in sites:
            pos = np.random.uniform(t.interval.left, t.interval.right)
        sites.add(pos)
        site = tables.sites.add_row(pos, ancestral_state='0')
        mtime = np.floor(np.random.uniform(time, ptime))
        assert mtime >= time and mtime < ptime
        md = {'s': 1e-3,
              'h': 1.0,
              'origin': int(mtime),
              'label': np.uint16(0),
              'neutral': 0,
              'key': np.uint64(0),
              }
        tables.mutations.add_row(site, n, '1', time=mtime, metadata=md)
    tables.sort()
    ts = tables.tree_sequence()
    pop = fwdpy11.DiploidPopulation.create_from_tskit(
        ts, import_mutations=True)


def test_complex_start_stop_workflow():
    yaml = """
description: split
time_units: generations
demes:
 - name: A
   epochs:
    - start_size: 10
      end_time: 10
 - name: B
   start_time: 10
   ancestors: [A]
   epochs:
    - start_size: 10
      end_size: 50
 - name: C
   start_time: 10
   ancestors: [A]
   epochs:
    - start_size: 10
      end_size: 50
"""
    demography = fwdpy11.ForwardDemesGraph.from_demes(
        yaml, burnin=10, burnin_is_exact=True)
    pdict = {'recregions': [],
             'nregions': [],
             'sregions': [],
             'rates': (0, 0, 0),
             'gvalue': fwdpy11.Multiplicative(2.),
             'simlen': 11,
             'demography': demography
             }
    params = fwdpy11.ModelParams(**pdict)
    pop = fwdpy11.DiploidPopulation(demography.initial_sizes, 1.0)
    rng = fwdpy11.GSLrng(10)
    fwdpy11.evolvets(rng, pop, params, 10)
    assert len(pop.deme_sizes()) == 2
    assert pop.generation == 11
    ts = pop.dump_tables_to_tskit()
    restored = fwdpy11.DiploidPopulation.create_from_tskit(ts)
    pdict['simlen'] = 100
    params = fwdpy11.ModelParams(**pdict)
    # Evolve rest of model
    fwdpy11.evolvets(rng, restored, params, 10)
    model_time = restored.generation
    assert model_time == 20
    ts = restored.dump_tables_to_tskit()
    restored2 = fwdpy11.DiploidPopulation.create_from_tskit(ts)
    fwdpy11.evolvets(rng, restored2, params, 10)
    assert model_time == restored2.generation



from typing import Dict, Optional, Union

import demes

from fwdpy11._types.demographic_model_citation import DemographicModelCitation
from fwdpy11._types.demographic_model_details import DemographicModelDetails
from fwdpy11._types.forward_demes_graph import ForwardDemesGraph


def demography_from_demes(
    dg: Union[str, demes.Graph], burnin: int,
    round_non_integer_sizes=Optional[bool],
) -> DemographicModelDetails:
    """
    The deme graph, dg, can be either a string or a resolved deme-graph.

    :param dg: A demes graph, either the object itself or a string for the YAML
        specified demography.
    :param int burnin: The factor for the burnin time, so that burn in occurs for
        burnin * N generations.

    .. versionadded:: 0.14.0
    """
    if isinstance(dg, str):
        g = demes.load(dg)
        source = {"demes_yaml_file": dg}
    else:
        g = dg
        source = None

    if not isinstance(burnin, int):
        raise ValueError("The burn in factor must be an integer")
    if burnin < 0:
        raise ValueError("Burn in factor must be non-negative")

    fg = ForwardDemesGraph.from_demes(
        g, burnin, round_non_integer_sizes=round_non_integer_sizes)

    demography = _build_from_foward_demes_graph(fg, burnin, source)
    return demography


def _build_from_foward_demes_graph(
    fg: ForwardDemesGraph, burnin: int, source: Optional[Dict] = None
) -> DemographicModelDetails:
    """
    The workhorse.
    """
    idmap = fg.deme_labels

    if fg.graph.doi != "None":
        doi = fg.graph.doi
    else:
        doi = None

    _initial_sizes = {i: j for i, j in enumerate(
        fg._parental_deme_sizes_at_time_zero()) if j > 0}
    return DemographicModelDetails(
        model=fg,
        name=fg.graph.description,
        source=source,
        parameters=None,
        citation=DemographicModelCitation(
            DOI=doi, full_citation=None, metadata=None),
        metadata={
            "deme_labels": idmap,
            "initial_sizes": _initial_sizes,
            "burnin_time": fg.burnin_generation,
            "total_simulation_length": fg.final_generation
        },
    )



#
# Copyright (C) 2017-2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import math

import fwdpy11
import fwdpy11.custom_genetic_value_decorators


@fwdpy11.custom_genetic_value_decorators.default_update
class PyAdditiveGSS(fwdpy11.PyDiploidGeneticValue):
    def __init__(self, opt, VS):
        self.opt = opt
        self.VS = VS
        fwdpy11.PyDiploidGeneticValue.__init__(self, 1, None, None)

    def genetic_value_to_fitness(
        self, data: fwdpy11.DiploidGeneticValueToFitnessData
    ) -> float:
        return math.e ** (
            -((data.offspring_metadata.g + data.offspring_metadata.e - self.opt) ** 2)
            / (2 * self.VS)
        )

    def calculate_gvalue(self, data: fwdpy11.PyDiploidGeneticValueData) -> float:
        s = fwdpy11.strict_additive_effects(data.pop, data.offspring_metadata)
        memoryview(data)[0] = s
        return s



#
# Copyright (C) 2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

"""
This module contains various things to help
interact with ``tskit``.

.. versionadded:: 0.8.0
"""

import typing

import numpy as np
import tskit  # type: ignore

from ._flags import *  # NOQA
from .metadata import (  # NOQA
    DiploidMetadata,
    decode_individual_metadata,
    decode_mutation_metadata,
)


def get_toplevel_metadata(ts: tskit.TreeSequence,
                          name: str) -> typing.Optional[object]:
    """
    Extract top-level metadata from a tree sequence.

    :param ts: A tree sequence
    :type ts: tskit.TreeSequence
    :param name: Metadata field name
    :type name: str
    :returns: The metadata field, if `name` is a valid field, or
              `None` otherwise.
    """
    if name in ts.metadata:
        return ts.metadata[name]
    return None


def iterate_timepoints_with_individuals(
    ts, *, decode_metadata=False
):
    """
    Return an iterator over all unique time points with individuals.

    :param ts: A tree sequence
    :type ts: tskit.TreeSequence
    :param decode_metadata: Flag to decode individual metadata or not
    :type decode_metadata: bool

    For each time point a tuple of (time, nodes, metadata) is yielded.

    :param decode_individual_metadata: Whether to return decoded metadata.
    :type decode_individual_metadata: bool

    If `decode_individual_metadata` is `True`, metadata will be stored in
    a :class:`list` of :class:`fwdpy11.tskit_tools.DiploidMetadata`.
    If `False`, `None` will be yielded.
    """

    nodes_in_individuals = []
    node_times = []
    for individual in ts.individuals():
        for node in individual.nodes:
            nodes_in_individuals.append(node)
            node_times.append(ts.node(node).time)

    node_times = np.array(node_times)
    nodes_in_individuals = np.array(nodes_in_individuals)

    unique_node_times = np.unique(node_times)

    for utime in unique_node_times[::-1]:
        # Get the node tables rows in individuals at this time
        x = np.where(node_times == utime)
        node_table_rows = nodes_in_individuals[x]
        assert np.all(
            np.array([ts.node(i).time for i in node_table_rows]) == utime)

        # Get the individuals
        individuals = np.unique(
            np.array([ts.node(i).individual for i in node_table_rows])
        )
        assert not np.any(individuals == tskit.NULL)

        if decode_metadata is True:
            # now, let's decode the individual metadata for this time slice
            decoded_individual_metadata = decode_individual_metadata(
                ts,
                individuals,
            )
        else:
            decoded_individual_metadata = None
        yield utime, node_table_rows, decoded_individual_metadata



#
# Copyright (C) 2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import unittest

import fwdpy11
import msprime
import numpy as np

# NOTE: these tests can all be improved in the future
# by having a conversion from msprime/tskit that lifts
# over mutations assuming that they are neutral.
# With that in place, we can compare to the msprime/tskit
# genotype matrix, which makes for a more independent test
# of correctness.


def fs_from_ndarray(gm):
    gm_rc = np.sum(gm, axis=1)
    gm_uc = np.unique(gm_rc, return_counts=True)
    gm_fs = np.zeros(gm.shape[1] + 1, dtype=np.int32)
    gm_fs[gm_uc[0]] += gm_uc[1]
    return gm_fs


class TestSingleDemeCase(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        Ne = 1000
        Nr = 100.0
        ts = msprime.simulate(
            2 * Ne, Ne=Ne, recombination_rate=Nr / Ne, random_seed=666
        )
        self.pop = fwdpy11.DiploidPopulation.create_from_tskit(ts)
        rng = fwdpy11.GSLrng(12343)
        fwdpy11.infinite_sites(rng, self.pop, Nr / Ne)

    def test_comparison_to_genotype_matrix(self):
        dm = fwdpy11.data_matrix_from_tables(
            self.pop.tables,
            self.pop.alive_nodes,
            record_neutral=True,
            record_selected=False,
        )
        gm = np.array(dm.neutral, copy=False)
        gm_fs = fs_from_ndarray(gm)
        tc_fs = self.pop.tables.fs([self.pop.alive_nodes])
        self.assertTrue(np.array_equal(gm_fs[1:-1], tc_fs.data[1:-1]))
        self.assertTrue(np.array_equal(gm_fs, tc_fs.data))

    def test_comparison_to_genotype_matrix_for_sample(self):
        A, B = 103, 210
        dm = fwdpy11.data_matrix_from_tables(
            self.pop.tables,
            self.pop.alive_nodes[A:B],
            record_neutral=True,
            record_selected=False,
        )
        gm = np.array(dm.neutral, copy=False)
        gm_fs = fs_from_ndarray(gm)
        tc_fs = self.pop.tables.fs([self.pop.alive_nodes[A:B]])
        self.assertTrue(np.array_equal(gm_fs[1:-1], tc_fs.data[1:-1]))

    def test_skipping_neutral_variants(self):
        tc_fs = self.pop.tables.fs([self.pop.alive_nodes], include_neutral=False)
        self.assertEqual(tc_fs.sum(), 0)

    def test_empty_samples_list(self):
        with self.assertRaises(ValueError):
            self.pop.tables.fs([])

    def test_nodes_out_of_range(self):
        with self.assertRaises(ValueError):
            samples = np.array([len(self.pop.tables.nodes)], dtype=np.int32)
            self.pop.tables.fs([samples])

    def test_two_overlapping_windows(self):
        tc_fs = self.pop.tables.fs([self.pop.alive_nodes])
        wfs = self.pop.tables.fs(
            [self.pop.alive_nodes], windows=[(0, 1.0 / 3.0), (1.0 / 3.0, 1.0)]
        )
        self.assertTrue(np.array_equal(wfs, tc_fs))

    def test_three_overlapping_windows(self):
        tc_fs = self.pop.tables.fs([self.pop.alive_nodes])
        wfs = self.pop.tables.fs(
            [self.pop.alive_nodes],
            windows=[(0, 1.0 / 3.0), (1.0 / 3.0, 2.0 / 3.0), (2.0 / 3.0, 1)],
        )
        self.assertTrue(np.array_equal(wfs, tc_fs))

    def test_random_number_of_windows(self):
        nw = np.random.randint(100, 300)
        lefts = np.arange(nw) / (nw - 1)
        windows = [(lefts[i], lefts[i + 1]) for i in range(len(lefts) - 1)]
        tc_fs = self.pop.tables.fs([self.pop.alive_nodes])
        wfs = self.pop.tables.fs([self.pop.alive_nodes], windows=windows)
        self.assertTrue(np.array_equal(wfs, tc_fs))

    def test_separated_windows(self):
        windows = [(0, 0.25), (0.66, 0.9)]
        dm = fwdpy11.data_matrix_from_tables(
            self.pop.tables,
            self.pop.alive_nodes,
            record_neutral=True,
            record_selected=False,
        )
        gm = np.array(dm.neutral, copy=False)
        gm_pos = np.array([self.pop.mutations[k].pos for k in dm.neutral_keys])
        gm_pos_in_windows = np.where(
            (gm_pos < 0.25) | ((gm_pos >= 0.66) & (gm_pos < 0.9))
        )[0]
        gm = gm[gm_pos_in_windows, :]
        gm_fs = fs_from_ndarray(gm)
        tc_fs = self.pop.tables.fs([self.pop.alive_nodes], windows=windows)
        self.assertTrue(np.array_equal(gm_fs, tc_fs))


class TestTwoDemeCase(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        Ne = 1000
        Nr = 100.0
        nodes_per_deme = 1000
        config = [
            msprime.PopulationConfiguration(nodes_per_deme),
            msprime.PopulationConfiguration(nodes_per_deme),
        ]

        events = [msprime.MassMigration(1 * Ne, 1, 0, 1.0)]
        ts = msprime.simulate(
            population_configurations=config,
            demographic_events=events,
            Ne=Ne,
            recombination_rate=Nr / Ne,
            random_seed=666,
        )
        self.pop = fwdpy11.DiploidPopulation.create_from_tskit(ts)
        rng = fwdpy11.GSLrng(12343)
        fwdpy11.infinite_sites(rng, self.pop, Nr / Ne)

    def test_marginal_deme_fs(self):
        a = self.pop.alive_nodes
        nodes = np.array(self.pop.tables.nodes, copy=False)
        d0 = a[np.where(nodes["deme"][a] == 0)[0]]
        d1 = a[np.where(nodes["deme"][a] == 1)[0]]

        for samples in (d0, d1):
            tc_fs = self.pop.tables.fs([samples])
            dm = fwdpy11.data_matrix_from_tables(
                self.pop.tables, samples, record_neutral=True, record_selected=True
            )
            gm = np.array(dm.neutral, copy=False)
            gm_fs = fs_from_ndarray(gm)
            self.assertTrue(np.array_equal(gm_fs[1:-1], tc_fs.data[1:-1]))

    def test_joint_deme_fs(self):
        a = self.pop.alive_nodes
        nodes = np.array(self.pop.tables.nodes, copy=False)
        d0 = a[np.where(nodes["deme"][a] == 0)[0]]
        d1 = a[np.where(nodes["deme"][a] == 1)[0]]

        tc_fs = self.pop.tables.fs([d0, d1])
        tc_fs0 = np.asarray(tc_fs.sum(axis=1).flatten())[0]
        tc_fs1 = np.asarray(tc_fs.sum(axis=0).flatten())[0]
        for i, j in zip((tc_fs0, tc_fs1), (d0, d1)):
            dm = fwdpy11.data_matrix_from_tables(
                self.pop.tables, j, record_neutral=True, record_selected=True
            )
            gm = np.array(dm.neutral, copy=False)
            gm_fs = fs_from_ndarray(gm)
            self.assertTrue(np.array_equal(gm_fs[1:-1], i.data[1:-1]))

    def test_joint_deme_fs_marginalize(self):
        a = self.pop.alive_nodes
        nodes = np.array(self.pop.tables.nodes, copy=False)
        d0 = a[np.where(nodes["deme"][a] == 0)[0]]
        d1 = a[np.where(nodes["deme"][a] == 1)[0]]
        tc_fs = self.pop.tables.fs([d0, d1])
        tc_fs_deme0 = np.asarray(tc_fs.sum(axis=1).flatten())[0]
        tc_fs_deme1 = np.asarray(tc_fs.sum(axis=0).flatten())[0]

        tc_fs2 = self.pop.tables.fs([d0, d1], marginalize=True)
        self.assertTrue(np.array_equal(tc_fs_deme0, tc_fs2[0]))
        self.assertTrue(np.array_equal(tc_fs_deme1, tc_fs2[1]))


if __name__ == "__main__":
    unittest.main()
